# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for trax.data.tf_inputs."""

import collections
import os
from unittest import mock

import gin
import numpy as np
from t5.data import assert_dataset
from t5.data import preprocessors as t5_processors
import tensorflow as tf
import tensorflow_datasets as tfds
from trax.data import inputs  # pylint: disable=unused-import
from trax.data import tf_inputs

pkg_dir, _ = os.path.split(__file__)
_TESTDATA = os.path.join(pkg_dir, 'testdata')


def _test_dataset_ints(inp_lengths, tgt_lengths):
  """Create a test dataset of int64 tensors of given shapes."""

  def generator():
    for inp_len, tgt_len in zip(inp_lengths, tgt_lengths):
      inp = np.ones([inp_len], dtype=np.int64)
      tgt = np.ones([tgt_len], dtype=np.int64)
      yield {'inputs': inp, 'targets': tgt}

  types = {'inputs': tf.int64, 'targets': tf.int64}
  shapes = {'inputs': tf.TensorShape([None]), 'targets': tf.TensorShape([None])}
  return tf.data.Dataset.from_generator(
      generator, output_types=types, output_shapes=shapes)


def _load_dataset(name, split='train'):
  return tfds.load(
      name=name, split=split, data_dir=_TESTDATA, shuffle_files=False)


def _c4_dataset(split='train'):
  return _load_dataset('c4:2.3.0', split=split)


def _spm_path():
  return os.path.join(_TESTDATA, 'sentencepiece.model')


def _t5_gin_config():
  # The following pages worth of gin configuration are required because a lot
  # of T5 functions have `gin.REQUIRED` in code, i.e. you cannot use these
  # functions at all without having configured gin.

  noise_density = 0.15
  max_input_length = 50

  # What preprocessors to apply - we select a random chunk of the document if
  # it exceeds a certain lengths (`select_random_chunk`), then split up long
  # examples (`split_tokens`) and finally the denoising objective (`denoise`).
  #
  # In addition to this T5 concates multiple documents together to reduce
  # padding (`reduce_concat_tokens`) after `select_random_chunk`, but we skip
  # that since we don't do sequence packing.
  gin.bind_parameter('unsupervised.preprocessors', [
      t5_processors.select_random_chunk,
      t5_processors.split_tokens,
      t5_processors.denoise,
  ])

  # select_random_chunk
  gin.bind_parameter('select_random_chunk.feature_key', 'targets')
  gin.bind_parameter('select_random_chunk.max_length', max_input_length)

  # reduce_concat_tokens
  gin.bind_parameter('random_spans_helper.extra_tokens_per_span_inputs', 1)
  gin.bind_parameter('random_spans_helper.extra_tokens_per_span_targets', 1)
  gin.bind_parameter('random_spans_helper.inputs_length', max_input_length)
  gin.bind_parameter('random_spans_helper.mean_noise_span_length', 3.0)
  gin.bind_parameter('random_spans_helper.noise_density', noise_density)

  # split_tokens
  gin.bind_parameter('split_tokens.max_tokens_per_segment',
                     t5_processors.random_spans_tokens_length())

  # denoise
  gin.bind_parameter('denoise.inputs_fn',
                     t5_processors.noise_span_to_unique_sentinel)
  gin.bind_parameter('denoise.noise_density', noise_density)
  gin.bind_parameter('denoise.noise_mask_fn',
                     t5_processors.random_spans_noise_mask)
  gin.bind_parameter('denoise.targets_fn',
                     t5_processors.nonnoise_span_to_unique_sentinel)


class TFInputsTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()
    gin.clear_config()


  def test_TFDS_single_host_with_eval_holdout(self):
    train_ds_gen = tf_inputs.TFDS(
        'c4/en:2.3.0',
        data_dir=_TESTDATA,
        train=True,
        host_id=0,
        keys=('text',),
        n_hosts=1,
        eval_holdout_size=0.1)

    # Just ensure that this doesn't crash.
    for d in train_ds_gen():
      print(f'Train: {d}')
      break

    valid_ds_gen = tf_inputs.TFDS(
        'c4/en:2.3.0',
        data_dir=_TESTDATA,
        train=False,
        host_id=0,
        keys=('text',),
        n_hosts=1,
        eval_holdout_size=0.1)

    # Just ensure that this doesn't crash.
    for d in valid_ds_gen():
      print(f'Eval: {d}')
      break

  def test_TFDS_single_host_with_eval_holdout_no_valid_split(self):
    train_ds_gen = tf_inputs.TFDS(
        'para_crawl/ende',
        data_dir=_TESTDATA,
        train=True,
        host_id=0,
        keys=('en', 'de'),
        n_hosts=1,
        eval_holdout_size=0.1)

    # Just ensure that this doesn't crash.
    for d in train_ds_gen():
      print(f'Train: {d}')
      break

    # para_crawl doesn't have a validation set, see that this still doesn't
    # crash because of eval_holdout_set.
    valid_ds_gen = tf_inputs.TFDS(
        'para_crawl/ende',
        data_dir=_TESTDATA,
        train=False,
        host_id=0,
        keys=('en', 'de'),
        n_hosts=1,
        eval_holdout_size=0.1)

    # Just ensure that this doesn't crash.
    for d in valid_ds_gen():
      print(f'Eval: {d}')
      break

  def test_TFDS_mnli_split_is_eval(self):
    with mock.patch('tensorflow_datasets.load') as tfds_load:
      with mock.patch('trax.data.tf_inputs.download_and_prepare',
                      lambda _, data_dir: data_dir):
        _ = tf_inputs.TFDS('glue/mnli',
                           keys=('premise', 'hypothesis'),
                           train=False)
      call_kwargs = tfds_load.call_args[1]
      self.assertEqual(call_kwargs['split'], 'validation_matched')

  def test_TFDS_mnli_split_is_alt_eval(self):
    with mock.patch('tensorflow_datasets.load') as tfds_load:
      with mock.patch('trax.data.tf_inputs.download_and_prepare',
                      lambda _, data_dir: data_dir):
        _ = tf_inputs.TFDS('glue/mnli',
                           keys=('premise', 'hypothesis'),
                           train=False,
                           use_alt_eval=True)
      call_kwargs = tfds_load.call_args[1]
      self.assertEqual(call_kwargs['split'], 'validation_mismatched')

  def test_convert_to_unicode(self):

    def dataset1():
      yield (b'Audentes fortuna iuvat.', b'Fortune favors the bold.')

    def dataset2():
      yield (b'\x81aabb', b'Value')

    convert_function1 = tf_inputs.ConvertToUnicode(keys=[0])
    convert_output1 = next(convert_function1(dataset1()))
    self.assertEqual(convert_output1[0], 'Audentes fortuna iuvat.')
    self.assertEqual(convert_output1[1], b'Fortune favors the bold.')
    self.assertIsInstance(convert_output1[0], str)
    self.assertIsInstance(convert_output1[1], bytes)

    # Contains an invalid bytes array from the point of view of UTF-8.
    try:
      convert_function2 = tf_inputs.ConvertToUnicode(keys=[0])
      convert_output2 = next(convert_function2(dataset2()))
    except UnicodeDecodeError:
      self.fail('ConvertToUnicode threw UnicodeDecodeError.')
    self.assertEqual(convert_output2[0], 'aabb')
    self.assertIsInstance(convert_output2[0], str)

  def test_tokenize_detokenize(self):

    def dataset():
      yield 'I have a cat.'

    # Character-level.
    tok_char = list(tf_inputs.tokenize(dataset(), vocab_type='char'))
    self.assertAllEqual(tok_char[0],
                        np.array([ord(c) for c in 'I have a cat.']))
    detok = tf_inputs.detokenize(tok_char[0], vocab_type='char')
    self.assertEqual(detok, 'I have a cat.')

    # Sentencepiece.
    tok_spc = list(
        tf_inputs.tokenize(
            dataset(),
            vocab_type='sentencepiece',
            vocab_dir=_TESTDATA,
            vocab_file='sentencepiece.model'))
    self.assertAllEqual(tok_spc[0], np.array([27, 43, 3, 9, 1712, 5]))
    detok = tf_inputs.detokenize(
        list(tok_spc[0]),
        vocab_type='sentencepiece',
        vocab_dir=_TESTDATA,
        vocab_file='sentencepiece.model')
    self.assertEqual(detok, 'I have a cat.')

    # Subword.
    tok_sbw = list(
        tf_inputs.tokenize(
            dataset(),
            vocab_type='subword',
            vocab_dir=_TESTDATA,
            vocab_file='en_8k.subword'))
    self.assertAllEqual(tok_sbw[0], np.array([139, 96, 12, 2217, 2, 21]))
    detok = tf_inputs.detokenize(
        tok_sbw[0],
        vocab_type='subword',
        vocab_dir=_TESTDATA,
        vocab_file='en_8k.subword')
    self.assertEqual(detok, 'I have a cat.')

    # bert-lowercase
    tok_sbw = list(
        tf_inputs.tokenize(
            dataset(),
            vocab_type='bert-lowercase',
            vocab_dir=_TESTDATA,
            vocab_file='bert_uncased_vocab.txt'))
    self.assertAllEqual(tok_sbw[0], np.array([1045, 2031, 1037, 4937, 1012]))
    detok = tf_inputs.detokenize(
        tok_sbw[0],
        vocab_type='bert-lowercase',
        vocab_dir=_TESTDATA,
        vocab_file='bert_uncased_vocab.txt')
    self.assertEqual(detok, 'i have a cat .')
    # note: BERT tokenizer is not reversible, therefore
    # difference between original input

  def test_tokenize_keys_reservedids(self):

    def dataset():
      yield ('Cat.', 'Dog.')

    tok_char1 = list(
        tf_inputs.tokenize(dataset(), vocab_type='char', n_reserved_ids=5))
    self.assertAllEqual(tok_char1[0][0], np.array([ord(c) + 5 for c in 'Cat.']))
    self.assertAllEqual(tok_char1[0][1], np.array([ord(c) + 5 for c in 'Dog.']))

    tok_char2 = list(
        tf_inputs.tokenize(
            dataset(), keys=[0], vocab_type='char', n_reserved_ids=2))
    self.assertAllEqual(tok_char2[0][0], np.array([ord(c) + 2 for c in 'Cat.']))
    self.assertEqual(tok_char2[0][1], 'Dog.')

  def test_tokenize_dict(self):

    def dataset():
      yield {'a': 'Cat.', 'b': 'Dog.'}

    tok_char1 = list(tf_inputs.tokenize(dataset(), vocab_type='char'))
    self.assertAllEqual(tok_char1[0]['a'], np.array([ord(c) for c in 'Cat.']))
    self.assertAllEqual(tok_char1[0]['b'], np.array([ord(c) for c in 'Dog.']))

    tok_char2 = list(
        tf_inputs.tokenize(dataset(), keys=['a'], vocab_type='char'))
    self.assertAllEqual(tok_char2[0]['a'], np.array([ord(c) for c in 'Cat.']))
    self.assertEqual(tok_char2[0]['b'], 'Dog.')

  def test_vocab_size(self):
    # Character-level.
    char_size = tf_inputs.vocab_size(vocab_type='char', n_reserved_ids=11)
    self.assertEqual(char_size, 256 + 11)
    # Sentencepiece.
    spc_size = tf_inputs.vocab_size(
        vocab_type='sentencepiece',
        vocab_dir=_TESTDATA,
        vocab_file='sentencepiece.model')
    self.assertEqual(spc_size, 32000)
    # Subword.
    sbw_size = tf_inputs.vocab_size(
        vocab_type='subword', vocab_dir=_TESTDATA, vocab_file='en_8k.subword')
    self.assertEqual(sbw_size, 8183)
    # Bert_uncased.
    sbw_size = tf_inputs.vocab_size(
        vocab_type='bert-lowercase',
        vocab_dir=_TESTDATA,
        vocab_file='bert_uncased_vocab.txt')
    self.assertEqual(sbw_size, 30522)

  def test_c4_bare_preprocess_fn(self):
    dataset = _c4_dataset()

    example = list(tfds.as_numpy(dataset.take(1)))[0]

    # Targets are NOT in the example.
    self.assertNotIn('targets', example)
    self.assertIn('text', example)
    text = example['text']

    # This should convert the dataset to an inputs/targets that are tokenized.
    dataset = tf_inputs.c4_bare_preprocess_fn(dataset, spm_path=_spm_path())

    example = list(tfds.as_numpy(dataset.take(1)))[0]

    # Earlier text is now stored in targets_pretokenized
    self.assertIn('targets_pretokenized', example)
    self.assertEqual(example['targets_pretokenized'], text)

    # Targets are now tokenized.
    self.assertIn('targets', example)
    self.assertIsInstance(example['targets'], np.ndarray)
    self.assertEqual(example['targets'].dtype, np.int64)
    self.assertGreater(len(example['targets']), 0)
    self.assertEqual(example['targets'][-1], 1)  # we add EOS at the end.

    # Inputs exist but is empty because t5 preprocessors' unsupervised wasn't
    # gin configured with any.
    self.assertIn('inputs', example)
    self.assertEqual(len(example['inputs']), 0)

  def test_c4_preprocess(self):

    def load_c4_dataset(split='train'):
      dataset = _c4_dataset(split=split)
      return dataset.map(lambda example: (example, example['text']))

    def examine_processed_dataset(proc_dataset):
      count = 0
      lengths = []
      for example in tfds.as_numpy(proc_dataset):
        count += 1
        ex = example[0]
        # Targets are in the example.
        self.assertIn('targets', ex)
        self.assertEqual(ex['targets'].dtype, np.int64)
        lengths.append(len(ex['targets']))
      return count, lengths

    unfiltered_count = 0
    for example in tfds.as_numpy(load_c4_dataset()):
      unfiltered_count += 1
      # Targets are NOT in the example.
      self.assertNotIn('targets', example[0])

    proc_dataset = tf_inputs.c4_preprocess(load_c4_dataset(), False, 2048)

    # `examine_processed_dataset` has some asserts in it.
    proc_count, char_lengths = examine_processed_dataset(proc_dataset)

    # Both the original and filtered datasets have examples.
    self.assertGreater(unfiltered_count, 0)
    self.assertGreater(proc_count, 0)

    # Because we filter out some entries on length.
    self.assertLess(proc_count, unfiltered_count)

    # Preprocess using the sentencepiece model in testdata.
    spc_proc_dataset = tf_inputs.c4_preprocess(
        load_c4_dataset(),
        False,
        2048,
        tokenization='spc',
        spm_path=_spm_path())

    spc_proc_count, spc_lengths = examine_processed_dataset(spc_proc_dataset)

    # spc shortens the target sequence a lot, should be almost equal to
    # unfiltered
    self.assertLessEqual(proc_count, spc_proc_count)
    self.assertEqual(unfiltered_count, spc_proc_count)

    # Assert all spc_lengths are lesser than their char counterparts.
    for spc_len, char_len in zip(spc_lengths, char_lengths):
      self.assertLessEqual(spc_len, char_len)

  def test_c4(self):
    gin.bind_parameter('c4_preprocess.max_target_length', 2048)
    gin.bind_parameter('c4_preprocess.tokenization', 'spc')
    gin.bind_parameter('c4_preprocess.spm_path', _spm_path())

    # Just make sure this doesn't throw.
    _ = tf_inputs.data_streams(
        'c4',
        data_dir=_TESTDATA,
        input_name='targets',
        target_name='text',
        preprocess_fn=tf_inputs.c4_preprocess)

  def test_c4_bare_preprocess_fn_denoising_objective(self):
    _t5_gin_config()

    dataset = _c4_dataset()
    dataset = tf_inputs.c4_bare_preprocess_fn(dataset, spm_path=_spm_path())

    example = list(tfds.as_numpy(dataset.take(1)))[0]

    # Assertions now.

    self.assertIn('targets', example)
    targets = example['targets']
    self.assertIsInstance(targets, np.ndarray)
    self.assertEqual(targets.dtype, np.int64)
    self.assertGreater(len(targets), 0)

    self.assertIn('inputs', example)
    _inputs = example['inputs']  # pylint: disable=invalid-name
    self.assertIsInstance(_inputs, np.ndarray)
    self.assertEqual(_inputs.dtype, np.int64)
    self.assertGreater(len(_inputs), 0)

    # WHP inputs will have the bulk of the text.
    self.assertGreater(len(_inputs), len(targets))

    # WHP there will be one sentinel token in the inputs and targets.
    inputs_counter = collections.Counter(_inputs.tolist())
    targets_counter = collections.Counter(targets.tolist())
    self.assertEqual(1, inputs_counter[31999])
    self.assertEqual(1, targets_counter[31999])

  def test_c4_pretrain(self):
    _t5_gin_config()

    gin.bind_parameter('c4_bare_preprocess_fn.spm_path', _spm_path())

    gin.bind_parameter('batcher.batch_size_per_device', 8)
    gin.bind_parameter('batcher.eval_batch_size', 8)
    gin.bind_parameter('batcher.max_eval_length', 50)
    gin.bind_parameter('batcher.buckets', ([51], [8, 1]))

    # Just make sure this doesn't throw.
    _ = tf_inputs.data_streams(
        'c4',
        data_dir=_TESTDATA,
        input_name='inputs',
        target_name='targets',
        bare_preprocess_fn=tf_inputs.c4_bare_preprocess_fn)

  def test_generic_text_dataset_preprocess_fn(self):
    dataset = _load_dataset('squad/v1.1:3.0.0')

    example, = tfds.as_numpy(dataset.take(1))

    self.assertNotIn('inputs', example)
    self.assertNotIn('targets', example)

    proc_dataset = tf_inputs.generic_text_dataset_preprocess_fn(
        dataset,
        spm_path=_spm_path(),
        text_preprocess_fns=[lambda ds, training: t5_processors.squad(ds)],
        copy_pretokenized=True,
        debug_print_examples=True,
        debug_print_examples_rate=1.0)

    proc_example, = tfds.as_numpy(proc_dataset.take(1))

    self.assertIn('inputs', proc_example)
    self.assertIn('targets', proc_example)

    self.assertEqual(proc_example['inputs'].dtype, np.int32)
    self.assertEqual(proc_example['targets'].dtype, np.int32)

  # TODO(afrozm): Why does this test take so much time?
  def test_inputs_using_generic_text_dataset_preprocess_fn(self):
    gin.bind_parameter('generic_text_dataset_preprocess_fn.spm_path',
                       _spm_path())
    gin.bind_parameter('generic_text_dataset_preprocess_fn.text_preprocess_fns',
                       [lambda ds, training: t5_processors.squad(ds)])

    # Just make sure this doesn't throw.
    def data_streams():
      return tf_inputs.data_streams(
          'squad',
          data_dir=_TESTDATA,
          input_name='inputs',
          target_name='targets',
          bare_preprocess_fn=tf_inputs.generic_text_dataset_preprocess_fn,
          shuffle_buffer_size=1)

    n_devices = 3

    squad_inputs = inputs.batcher(
        data_streams=data_streams,
        max_eval_length=512,
        buckets=([
            513,
        ], [n_devices, n_devices]))

    eval_stream = squad_inputs.eval_stream(n_devices)
    inps, tgts, _ = next(eval_stream)

    # We can only assert that the batch dim gets divided by n_devices.
    self.assertEqual(inps.shape[0] % n_devices, 0)
    self.assertEqual(tgts.shape[0] % n_devices, 0)

  def test_filter_dataset_on_len(self):
    # {1, 2}, {2, 4}, {3, 6} ... {10, 20}
    ds = _test_dataset_ints(range(1, 11), range(2, 21, 2))

    ds1 = tf_inputs.filter_dataset_on_len(ds, True, {
        'inputs': [4, 8],
        'targets': [14, 20]
    })
    # Only {7, 14} and {8, 16} satisfy this.
    self.assertLen(list(ds1.as_numpy_iterator()), 2)

    ds2 = tf_inputs.filter_dataset_on_len(
        ds,
        False,
        len_map={
            'inputs': [4, 8],
            'targets': [14, 20]
        },
        filter_on_eval=False)
    # This is eval and we aren't supposed to filter it.
    self.assertLen(list(ds2.as_numpy_iterator()), 10)

    ds3 = tf_inputs.filter_dataset_on_len(
        ds,
        False,
        len_map={
            'inputs': [4, 8],
            'targets': [14, 20]
        },
        filter_on_eval=True)
    # This is eval and we are asked to filter it.
    self.assertLen(list(ds3.as_numpy_iterator()), 2)

  def test_truncate_dataset_on_len(self):
    ds = _test_dataset_ints([5, 6, 7], [8, 9, 10])
    ds1 = tf_inputs.truncate_dataset_on_len(
        ds, True, len_map={
            'inputs': 6,
            'targets': 4
        })
    expected_ds = _test_dataset_ints([5, 6, 6], [4, 4, 4])

    # training, should filter.
    assert_dataset(ds1, list(expected_ds.as_numpy_iterator()))

    # not Training, shouldn't filter.
    ds2 = tf_inputs.truncate_dataset_on_len(
        ds, False, len_map={
            'inputs': 6,
            'targets': 4
        })
    assert_dataset(ds2, list(ds.as_numpy_iterator()))

    # not Training, but asked to filter, should filter.
    ds3 = tf_inputs.truncate_dataset_on_len(
        ds, False, len_map={
            'inputs': 6,
            'targets': 4
        }, truncate_on_eval=True)
    assert_dataset(ds3, list(expected_ds.as_numpy_iterator()))

  def test_get_t5_preprocessor_by_name(self):
    gin.clear_config()

    gin.parse_config("""
      get_t5_preprocessor_by_name.name = 'rekey'
      get_t5_preprocessor_by_name.fn_kwargs = {'key_map': {'inputs': 'other', 'targets': 'text'}}
    """)
    prep_rekey = tf_inputs.get_t5_preprocessor_by_name()
    og_dataset = tf.data.Dataset.from_tensors({
        'text': 'That is good.',
        'other': 'That is bad.'
    })
    training = True
    dataset = prep_rekey(og_dataset, training)
    assert_dataset(dataset, {
        'inputs': 'That is bad.',
        'targets': 'That is good.'
    })

  def test_pad_dataset_to_length(self):
    ds = _test_dataset_ints([5, 6, 7], [6, 7, 8])
    ds1 = tf_inputs.pad_dataset_to_length(
        ds, True, len_map={
            'inputs': 7,
            'targets': 10
        })

    expected_ds = [
        {
            'inputs': np.array([1, 1, 1, 1, 1, 0, 0], dtype=np.int64),
            'targets': np.array([1, 1, 1, 1, 1, 1, 0, 0, 0, 0], dtype=np.int64),
        },
        {
            'inputs': np.array([1, 1, 1, 1, 1, 1, 0], dtype=np.int64),
            'targets': np.array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0], dtype=np.int64),
        },
        {
            'inputs': np.array([1, 1, 1, 1, 1, 1, 1], dtype=np.int64),
            'targets': np.array([1, 1, 1, 1, 1, 1, 1, 1, 0, 0], dtype=np.int64),
        },
    ]

    assert_dataset(ds1, expected_ds)

  def test_lm_token_preprocessing(self):
    ds = _test_dataset_ints([1, 2, 3], [3, 2, 1])
    ds1 = tf_inputs.lm_token_preprocessing(ds, True)

    # pylint: disable=bad-whitespace
    expected_ds = [
        {
            'inputs': np.array([1, 0, 1, 1, 1], dtype=np.int64),
            'targets': np.array([1, 0, 1, 1, 1], dtype=np.int64),
            'mask': np.array([0, 0, 1, 1, 1], dtype=np.int64),
        },
        {
            'inputs': np.array([1, 1, 0, 1, 1], dtype=np.int64),
            'targets': np.array([1, 1, 0, 1, 1], dtype=np.int64),
            'mask': np.array([0, 0, 0, 1, 1], dtype=np.int64),
        },
        {
            'inputs': np.array([1, 1, 1, 0, 1], dtype=np.int64),
            'targets': np.array([1, 1, 1, 0, 1], dtype=np.int64),
            'mask': np.array([0, 0, 0, 0, 1], dtype=np.int64),
        },
    ]
    # pylint: enable=bad-whitespace

    assert_dataset(ds1, expected_ds)

  def test_create_bert_inputs(self):
    inputs_sentences_1 = [np.array([100, 150, 200])]
    inputs_sentences_2 = [np.array([300, 500])]
    labels = [np.array(1)]

    create_inputs_1 = tf_inputs.CreateBertInputs(False)
    create_inputs_2 = tf_inputs.CreateBertInputs(True)
    for res in create_inputs_1(zip(inputs_sentences_1, labels)):
      values, segment_embs, _, label, weight = res
      self.assertAllEqual(values, np.array([101, 100, 150, 200, 102]))
      self.assertAllEqual(segment_embs, np.zeros(5))
      self.assertEqual(label, np.int64(1))
      self.assertEqual(weight, np.int64(1))

    for res in create_inputs_2(
        zip(inputs_sentences_1, inputs_sentences_2, labels)):
      values, segment_embs, _, label, weight = res
      self.assertAllEqual(values,
                          np.array([101, 100, 150, 200, 102, 300, 500, 102]))
      exp_segment = np.concatenate((np.zeros(5), np.ones(3)))
      self.assertAllEqual(segment_embs, exp_segment)
      self.assertEqual(label, np.int64(1))
      self.assertEqual(weight, np.int64(1))

  def test_mask_random_tokens(self):
    """Test only standard tokens.

    This test deals with sentences composed of two parts: [100 CLS tokens, 100
    chosen standard tokens]. CLS is the token that is added at the beginning of
    the sentence and there is only one token in standard scenario. It is never
    masked because it is not a part of the sentence.
    This tests whether mask_random_tokens will:
      - mask only standard tokens
      - mask expected number of tokens (15 percent candidates for masking)
    """
    cls_token = 101
    mask_token = 103
    example_standard_token = 1001
    test_case_row = np.array([cls_token] * 100 + [example_standard_token] * 100)
    test_case = [(test_case_row.copy(),)]

    out, original_tokens, token_weights = next(
        tf_inputs.mask_random_tokens(test_case))
    # test whether original tokens are unchanged
    self.assertAllEqual(test_case_row, original_tokens)

    self.assertEqual(1, token_weights.sum())
    self.assertEqual(
        15,
        (token_weights > 0).sum())  # we should have 15 candidates for masking

    # 101 is a special token, so only 1001 should be masked
    self.assertAllEqual(out[:100], test_case_row[:100])

    # Each candidate has 0.8 probability to be masked while others have 0, so
    # no more than 15 tokens with MASK
    self.assertLessEqual((out == mask_token).sum(), 15)

  def test_bert_next_sentence_prediction_inputs(self):
    stream = tf_inputs.BertNextSentencePredictionInputs(
        'c4/en:2.3.0', data_dir=_TESTDATA, train=False, shuffle_size=1)
    exp_sent1 = 'Police were called to the carriageway around 6.'
    exp_sent2 = 'I am sorry we did not see how lost and alone you felt.'
    sent1, sent2, label = next(stream())
    self.assertEqual(exp_sent1, sent1)
    self.assertEqual(exp_sent2, sent2)
    self.assertFalse(label)

  def test_process_single_mathqa_example_0(self):
    # This is the first problem in the MathQA dataset.
    example = {
        'Problem':
            "the banker ' s gain of a certain sum due 3 years hence at 10 % "
            'per annum is rs . 36 . what is the present worth ?',
        'Rationale':
            '"explanation : t = 3 years r = 10 % td = ( bg × 100 ) / tr = ( '
            '36 × 100 ) / ( 3 × 10 ) = 12 × 10 = rs . 120 td = ( pw × tr )'
            ' / 100 ⇒ 120 = ( pw × 3 × 10 ) / 100 ⇒ 1200 = pw × 3 pw = '
            '1200 / 3 = rs . 400 answer : option a"',
        'options':
            'a ) rs . 400 , b ) rs . 300 , c ) rs . 500 , d ) rs . 350 , e ) '
            'none of these',
        'correct':
            'a',
        'annotated_formula':
            'divide(multiply(const_100, divide(multiply(36, const_100), '
            'multiply(3, 10))), multiply(3, 10))',
        'linear_formula':
            'multiply(n2,const_100)|multiply(n0,n1)|divide(#0,#1)|multiply(#2,const_100)|divide(#3,#1)|',
        'category':
            'gain'
    }

    answer_num, python_result, python_program, list_op, list_num = tf_inputs.process_single_mathqa_example(
        example)
    self.assertEqual(answer_num,
                     400)  # we know it, because correct answer is a)
    self.assertEqual(python_result, [3600.0, 30.0, 120.0, 12000.0, 400.0])

    self.assertEqual(python_program, [
        't0 = n2 * 100.0', 't1 = n0 * n1', 't2 = t0 / t1', 't3 = t2 * 100.0',
        't4 = t3 / t1'
    ])
    self.assertEqual(list_op, [
        'multiply(n2,const_100)', 'multiply(n0,n1)', 'divide(#0,#1)',
        'multiply(#2,const_100)', 'divide(#3,#1)'
    ])
    self.assertEqual(list_num, [3.0, 10.0, 36.0])

  def test_process_single_mathqa_example_1(self):
    # This is the third problem in the MathQA dataset.
    example = {
        'Problem':
            'sophia finished 2 / 3 of a book . she calculated that she '
            'finished 90 more pages than she has yet to read . how long is her'
            ' book ?',
        'Rationale':
            'let xx be the total number of pages in the book , then she '
            'finished 23 ⋅ x 23 ⋅ x pages . then she has x − 23 ⋅ x = '
            '13 ⋅ xx − 23 ⋅ x = 13 ⋅ x pages left . 23 ⋅ x − 13 '
            '⋅ x = 9023 ⋅ x − 13 ⋅ x = 90 13 ⋅ x = 9013 ⋅ x = 90 x'
            ' = 270 x = 270 so the book is 270 pages long . answer : b',
        'options': 'a ) 229 , b ) 270 , c ) 877 , d ) 266 , e ) 281',
        'correct': 'b',
        'annotated_formula': 'divide(90, subtract(const_1, divide(2, 3)))',
        'linear_formula': 'divide(n0,n1)|subtract(const_1,#0)|divide(n2,#1)',
        'category': 'general'
    }

    answer_num, python_result, python_program, list_op, list_num = tf_inputs.process_single_mathqa_example(
        example)
    self.assertEqual(answer_num,
                     270)  # we know it, because correct answer is b)
    self.assertAllClose(
        python_result,
        [0.6666666666666666, 0.33333333333333337, 269.99999999999994])
    self.assertEqual(python_program,
                     ['t0 = n0 / n1', 't1 = 1.0 - t0', 't2 = n2 / t1'])
    self.assertEqual(list_op,
                     ['divide(n0,n1)', 'subtract(const_1,#0)', 'divide(n2,#1)'])
    self.assertEqual(list_num, [2.0, 3.0, 90.0])

  def test_process_single_mathqa_example_with_import(self):
    # This is a training MathQA problem which involve an import.
    example = {
        'Problem':
            'the length of a rectangular garden is three times its width . if '
            'the area of the rectangular garden is 588 square meters , then '
            'what is the width of the rectangular garden ?',
        'Rationale':
            '\"let x be the width of the garden . 3 x ^ 2 = 588 x ^ 2 = 196 x '
            '= 14 the answer is c .\"',
        'options':
            'a ) 12 , b ) 13 , c ) 14 , d ) 15 , e ) 16',
        'correct':
            'c',
        'annotated_formula':
            'sqrt(divide(588, const_3))',
        'linear_formula':
            'divide(n0,const_3)|sqrt(#0)|',
        'category':
            'geometry'
    }

    answer_num, python_result, python_program, list_op, list_num = tf_inputs.process_single_mathqa_example(
        example)
    self.assertEqual(answer_num, 14)  # we know it, because correct answer is c)
    self.assertAllClose(python_result, [196, 14])
    self.assertEqual(
        python_program,
        ['t0 = n0 / 3.0', 't1 = math.sqrt(max(0, t0))'])
    self.assertEqual(list_op, ['divide(n0,const_3)', 'sqrt(#0)'])
    self.assertEqual(list_num, [588])

    # Below we execute twice the Python program and once the DSL program.
    target_values = 'import math\n'
    problem = example['Problem']
    for i in range(len(list_num)):
      target_values += 'n{} = {}\n'.format(i, list_num[i])
      problem += ' n{} = {}'.format(i, list_num[i])
    target_values += '\n'.join(python_program[:-1])
    final_line = python_program[-1].split('=')[1]
    target_values += '\nanswer ={}'.format(final_line)
    var_dict = {}
    exec(target_values, globals(), var_dict)  # pylint: disable=exec-used
    self.assertAllClose(var_dict['answer'], 14)
    self.assertAllClose(
        tf_inputs.execute_mathqa_program(problem, target_values.split('\n')),
        14)
    self.assertAllClose(
        tf_inputs.execute_mathqa_dsl_program(problem,
                                             [example['linear_formula']]), 14)


  def test_sentencepiece_tokenize(self):
    def dataset():
      yield 'I have a cat.'

    examples = []
    for example in tf_inputs.sentencepiece_tokenize(dataset(), _spm_path()):
      examples.append(example)
    toks = list(examples[0])
    self.assertSequenceEqual([27, 43, 3, 9, 1712, 5], toks)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests of learning rate schedules."""

import math

from absl.testing import absltest

from trax.supervised import lr_schedules


class LRFunctionsTest(absltest.TestCase):

  def test_warmup(self):
    lr_fn = lr_schedules.warmup(9, .01)

    # Linear warm-up.
    self.assertAlmostEqual(.001, lr_fn(1))
    self.assertAlmostEqual(.002, lr_fn(2))
    self.assertAlmostEqual(.005, lr_fn(5))
    self.assertAlmostEqual(.009, lr_fn(9))

    # Constant thereafter.
    self.assertAlmostEqual(.01, lr_fn(10))
    self.assertAlmostEqual(.01, lr_fn(11))
    self.assertAlmostEqual(.01, lr_fn(20))
    self.assertAlmostEqual(.01, lr_fn(300))
    self.assertAlmostEqual(.01, lr_fn(4000))

  def test_constant(self):
    lr_fn = lr_schedules.constant(.02)
    self.assertEqual(.02, lr_fn(1))
    self.assertEqual(.02, lr_fn(20))
    self.assertEqual(.02, lr_fn(300))
    self.assertEqual(.02, lr_fn(4000))
    self.assertEqual(.02, lr_fn(50000))
    self.assertEqual(.02, lr_fn(600000))
    self.assertEqual(.02, lr_fn(7000000))
    self.assertEqual(.02, lr_fn(80000000))
    self.assertEqual(.02, lr_fn(900000000))

  def test_warmup_and_rsqrt_decay(self):
    lr_fn = lr_schedules.warmup_and_rsqrt_decay(24, .25)

    # Warm-up.
    self.assertAlmostEqual(.01, lr_fn(1))
    self.assertAlmostEqual(.02, lr_fn(2))
    self.assertAlmostEqual(.23, lr_fn(23))
    self.assertAlmostEqual(.24, lr_fn(24))

    # Reciprocal square-root decay.
    self.assertAlmostEqual(.25 * (5 / math.sqrt(25)), lr_fn(25))
    self.assertAlmostEqual(.25 * (5 / math.sqrt(26)), lr_fn(26))
    self.assertAlmostEqual(.25 * (5 / math.sqrt(27)), lr_fn(27))
    self.assertAlmostEqual(.25 * (5 / math.sqrt(300)), lr_fn(300))
    self.assertAlmostEqual(.25 * (5 / math.sqrt(4000)), lr_fn(4000))
    self.assertAlmostEqual(.25 * (5 / math.sqrt(50000)), lr_fn(50000))

  def test_cosine_sawtooth(self):
    tail_fn = lr_schedules._CosineSawtoothTail(180, min_value=.1)
    lr_fn = lr_schedules._BodyAndTail(.3, tail_start=0, tail_fn=tail_fn)

    # First cycle
    self.assertAlmostEqual(.29998477, lr_fn(1))
    self.assertAlmostEqual(.28660254, lr_fn(30))
    self.assertAlmostEqual(.25, lr_fn(60))
    self.assertAlmostEqual(.20, lr_fn(90))
    self.assertAlmostEqual(.15, lr_fn(120))
    self.assertAlmostEqual(.10001523, lr_fn(179))

    # Second cycle
    self.assertEqual(.3, lr_fn(180))
    self.assertAlmostEqual(.29998477, lr_fn(181))
    self.assertAlmostEqual(.28660254, lr_fn(210))
    self.assertAlmostEqual(.25, lr_fn(240))
    self.assertAlmostEqual(.20, lr_fn(270))
    self.assertAlmostEqual(.15, lr_fn(300))
    self.assertAlmostEqual(.10001523, lr_fn(359))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for RL."""

from unittest import mock
from absl.testing import absltest
import numpy as np

from trax import shapes
from trax.models import rl


class RLTest(absltest.TestCase):

  def test_policy_forward_shape(self):
    mock_dist = mock.MagicMock()
    mock_dist.n_inputs = 4
    model = rl.Policy(policy_distribution=mock_dist)
    x = np.ones((2, 3))
    _, _ = model.init(shapes.signature(x))
    y = model(x)
    self.assertEqual(y.shape, (2, 4))

  def test_value_forward_shape(self):
    model = rl.Value()
    x = np.ones((2, 3))
    _, _ = model.init(shapes.signature(x))
    y = model(x)
    self.assertEqual(y.shape, (2, 1))

  def test_policy_and_value_forward_shape(self):
    mock_dist = mock.MagicMock()
    mock_dist.n_inputs = 4
    model = rl.PolicyAndValue(policy_distribution=mock_dist)
    x = np.ones((2, 3))
    _, _ = model.init(shapes.signature(x))
    ys = model(x)
    self.assertEqual([y.shape for y in ys], [(2, 4), (2, 1)])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Timing tests for decoding."""

import copy
import functools
import gc
import os
import time
from jax import test_util  # pylint: disable=unused-import
from jax.config import config
import numpy as np
import psutil
from tensorflow.compat.v2 import test

from trax import fastmath
from trax import layers as tl
from trax import models
from trax import shapes
from trax.supervised import decoding


def _size_of_model(model):
  def _size(x):
    try:
      return x.size
    except Exception:  # pylint: disable=broad-except
      return 0
  sizes = fastmath.nested_map(_size, model.weights)
  total_size = sum(fastmath.tree_flatten(sizes))
  return total_size


def _recurrent_delete(w):
  if 'delete' in dir(w):
    # Object has a 'delete' method, so it is a DeviceArray or something similar,
    # so we want to delete it.
    w.delete()
  elif isinstance(w, (list, tuple)):
    for x in w:
      _recurrent_delete(x)
  elif isinstance(w, dict):
    for x in w.values():
      _recurrent_delete(x)
  else:
    raise ValueError('Unknown type encountered in weights: {}'.format(type(w)))


def _memory_usage():
  gc.collect()
  return psutil.Process(os.getpid()).memory_info().rss


class DecodingTimingTest(test.TestCase):

  def _terraformer_decoding_time(self, settings):
    # Garbage collection influences the timing, so we turn it off.
    gc.disable()
    max_len = 16

    def _self_attention_fn():
      return functools.partial(
          tl.SelfAttention,
          predict_drop_len=2 * max_len,
          predict_mem_len=2 * max_len)

    def _causal_attention_fn():
      attn_layer, attn_kwargs = settings['attn']
      return functools.partial(
          attn_layer,
          max_inference_length=2 * max_len, **attn_kwargs)

    if settings['model'] == 'terraformer':
      pred_model = models.ConfigurableTerraformer(
          mode='predict',
          d_model=settings['d_model'],
          d_ff=settings['d_ff'],
          dropout=0.1,
          max_len=max_len,
          n_heads=settings['n_heads'],
          n_encoder_layers=settings['encoder_layers'],
          n_decoder_layers=settings['decoder_layers'],
          encoder_attention_type=_self_attention_fn(),
          encoder_decoder_attention_type=_causal_attention_fn(),
          input_vocab_size=settings['vocab'],
          ff_sparsity=settings['ff_sparsity'],
          ff_use_sru=settings['ff_use_sru'],
          ff_dropout=0.1,
          # ff_chunk_size=1024,
          # attention_chunk_size=1,
          n_decoder_attention_layers=settings['attention_layers'],
          loss_sparsity=settings['loss_sparsity'],
          pos_axial_shape=None,
          use_bfloat16=True,
      )
    elif settings['model'] == 'transformer':
      pred_model = models.ConfigurableTransformer(
          mode='predict',
          d_model=settings['d_model'],
          d_ff=settings['d_ff'],
          dropout=0.1,
          max_len=max_len,
          n_heads=settings['n_heads'],
          n_encoder_layers=settings['encoder_layers'],
          n_decoder_layers=settings['decoder_layers'],
          # encoder_attention_type=_self_attention_fn(),
          encoder_decoder_attention_type=_causal_attention_fn(),
          input_vocab_size=settings['vocab'],
          ff_sparsity=settings['ff_sparsity'],
          ff_use_sru=settings['ff_use_sru'],
          # ff_dropout=0.1,
          # ff_chunk_size=1024,
          # attention_chunk_size=1,
          # n_decoder_attention_layers=settings['attention_layers'],
          loss_sparsity=settings['loss_sparsity'],
          pos_axial_shape=None,
          # enc_dec_attention_sparsity=settings['enc_dec_sparsity'],
          # use_bfloat16=True,
      )
    else:
      assert False
    # We put acceleration outside of autoregressive_sample_stream, because
    # we want to have a separate run (separate input) for model compilation.
    pred_model = tl.Accelerate(pred_model)

    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
    shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)
    pred_model.init(input_signature=(shape1l, shape11))
    original_state = copy.deepcopy(pred_model.state)

    inputs_warmup = np.zeros((1, max_len), dtype=np.int32)
    inputs = np.arange(max_len, dtype=np.int32).reshape(1, max_len)

    # This is a warm-up run, for compilation.
    result, current_time = [], time.time()
    elapsed_warmup_times = []
    for index, sample in zip(range(0, 4), decoding.autoregressive_sample_stream(
        pred_model, inputs_warmup, temperature=0.0, accelerate=False)):
      del index  # unused
      result.append(sample[:, None])  # to be sure that the result is computed

      current_time, start_time = time.time(), current_time
      elapsed_warmup_times.append(current_time - start_time)

    # This is a real decoding timing run that we measure.
    pred_model.state = original_state
    result, current_time = [], time.time()
    elapsed_times = []
    for index, sample in zip(range(12), decoding.autoregressive_sample_stream(
        pred_model, inputs, temperature=0.0, accelerate=False)):
      del index  # unused
      result.append(sample[:, None])  # to be sure that the result is computed

      current_time, start_time = time.time(), current_time
      elapsed_times.append(current_time - start_time)
    peak_memory = _memory_usage()

    if min(elapsed_times[2:]) * 2 < max(elapsed_times[2:]):
      print('WARNING! High variance found in elapsed times! Settings: {} ; '
            'elapsed times: {} ; Probably more warm-up steps should be used, '
            'or model size should be increased.'.format(settings,
                                                        elapsed_times))
    # Check resulting shapes.
    s = np.concatenate(result, axis=1)
    self.assertEqual(s.shape[0], 1)
    self.assertEqual(s.shape[1], 12)
    model_size = int(_size_of_model(pred_model))

    # We delete the model weights, because in some situations they won't be
    # deleted automatically.
    _recurrent_delete(pred_model.weights)
    gc.enable()
    return model_size, elapsed_times, peak_memory

  def test_autoregressive_sample_terraformer_timing(self):
    template_to_use = 'medium_transformer'

    settings_templates = {
        # full model
        # # 54B params
        # 'full_model': {
        #     'encoder_layers': 6, 'decoder_layers': 36, 'vocab': 32000,
        #     'attention_layers': 2,
        #     'd_ff': 64*1024, 'd_model': 96*96, 'n_heads': 96,
        #     'ff_use_sru': (1, 64), 'ff_sparsity': (256, 32),
        #     'loss_sparsity': 8,
        #     'attn': (tl.MultiplicativeConvCausalAttention,
        #              {'length_kernel_size': 3, 'sparsity': 64})},

        # 1/18 of model (1/6 of encoder, 1/18 of decoder, full vocab)
        # 4B params
        # 'big_terraformer': {
        #     'model': 'terraformer',
        #     'encoder_layers': 1, 'decoder_layers': 2, 'vocab': 32000,
        #     'attention_layers': 2,
        #     'd_ff': int(5/8 * 64*1024), 'd_model': 96*96, 'n_heads': 96,
        #     'ff_use_sru': 0, 'ff_sparsity': 0, 'loss_sparsity': 0,
        #     'attn': (tl.CausalAttention, {})},

        # 'big_transformer': {
        #     'model': 'transformer',
        #     'encoder_layers': 1, 'decoder_layers': 2, 'vocab': 32000,
        #     'attention_layers': 2,
        #     'd_ff': int(5/8 * 64*1024), 'd_model': 96*96, 'n_heads': 96,
        #     'ff_use_sru': 0, 'ff_sparsity': 0, 'loss_sparsity': 0,
        #     'attn': (tl.CausalAttention, {})},

        # medium model
        # 275M params (only decoder)
        'medium_transformer': {
            'model': 'transformer',
            'encoder_layers': 2, 'decoder_layers': 24, 'vocab': 32000,
            'attention_layers': 2,
            'd_ff': 4*1024, 'd_model': 1024, 'n_heads': 16,
            'ff_use_sru': 0, 'ff_sparsity': 0, 'loss_sparsity': 0,
            'attn': (tl.CausalAttention, {})},
        # 'medium_terraformer': {
        #     'model': 'terraformer',
        #     'encoder_layers': 2, 'decoder_layers': 24, 'vocab': 32000,
        #     'attention_layers': 2,
        #     'd_ff': 4*1024, 'd_model': 1024, 'n_heads': 16,
        #     'ff_use_sru': 0, 'ff_sparsity': 0, 'loss_sparsity': 0,
        #     'attn': (tl.CausalAttention, {})},

    }

    sweep_settings = {
        # 'big_transformer': [  # for big
        #     dict(), # baseline
        #     {'ff_sparsity': (256, 32)},  # + Sparse FF
        #     {'attn': (  # + Sparse QKV
        #         tl.MultiplicativeConvCausalAttention,
        #         {'length_kernel_size': 3, 'sparsity': 64}),
        #      'd_ff': 64*1024,
        #      },
        #     {'ff_sparsity': (256, 32),
        #      'attn': (  # + Sparse FF+QKV
        #         tl.MultiplicativeConvCausalAttention,
        #         {'length_kernel_size': 3, 'sparsity': 64}),
        #      'd_ff': 64*1024,
        #      },
        # ],

        'medium_transformer': [  # for medium
            dict(),  # baseline

            {'ff_sparsity': 64,
             'attn': (  # Sparse FF+QKV
                 tl.MultiplicativeConvCausalAttention,
                 {'length_kernel_size': 3, 'sparsity': 16}),
             'd_ff': 6*1024,
             },

            # {'ff_sparsity': 64,  # Sparse FF+QKV + Loss
            #  'attn': (
            #     tl.MultiplicativeConvCausalAttention,
            #     {'length_kernel_size': 3, 'sparsity': 16}),
            #  'd_ff': 6*1024,
            #  'loss_sparsity': 4,
            #  },

            # {'attn': (  # Sparse QKV
            #     tl.MultiplicativeConvCausalAttention,
            #     {'length_kernel_size': 3, 'sparsity': 16}),
            #  'd_ff': 6*1024,
            #  },
            # {'loss_sparsity': 4},  # Sparse Loss
            # {'ff_sparsity': 64},  # Sparse FF

            # {'ff_sparsity': 128},  # + Sparse FF 128

            # APPENDIX below

            # different loss layers
            # {'loss_sparsity': 8},
            # {'loss_sparsity': 2},
            # {'loss_sparsity': 0},
        ],

        #  'big_terraformer': [  # for big terraformer
        #      dict(), # baseline
        #      {'ff_sparsity': 64},  # + Sparse FF  / Sparse FF 64
        #      {'ff_sparsity': 64,
        #       'attn': (  # + Sparse FF+QKV
        #          tl.MultiplicativeConvCausalAttention,
        #          {'length_kernel_size': 3, 'sparsity': 16}),
        #       'd_ff': 6*1024,
        #       },
        #      {'ff_sparsity': 64,  # + Sparse FF+QKV+Loss
        #       'attn': (
        #          tl.MultiplicativeConvCausalAttention,
        #          {'length_kernel_size': 3, 'sparsity': 16}),
        #       'd_ff': 6*1024,
        #       'loss_sparsity': 4,
        #       },

        #         ],

        # 'medium_terraformer': [  # for medium terraformer
        #     {'ff_sparsity': 64,  # + Sparse FF+QKV+Loss
        #      'attn': (
        #         tl.MultiplicativeConvCausalAttention,
        #         {'length_kernel_size': 3, 'sparsity': 16}),
        #      'd_ff': 6*1024,
        #      'loss_sparsity': 4,
        #      },
        # ],
    }

    encoding_times = []
    decoding_times = []
    sizes = []
    memories = []
    messages = []
    for override_settings in sweep_settings[template_to_use]:
      settings = copy.deepcopy(settings_templates[template_to_use])
      settings.update(override_settings)

      init_memory = _memory_usage()
      size, elapsed_times, peak_memory = (
          self._terraformer_decoding_time(settings))

      # TODO(jaszczur): Why is elapsed_times[0] always small?
      encoding_time = elapsed_times[1]
      decoding_time_10 = sum(elapsed_times[2:])

      after_memory = _memory_usage()
      model_memory_gigabytes = (peak_memory-init_memory)/1024**3
      decoding_time_diff = (max(elapsed_times[2:]) - min(elapsed_times[2:])) / 2
      decoding_time_diff_percent = int(
          decoding_time_diff / np.mean(elapsed_times) * 100)
      message = (
          '\n\n'
          'Params: {}\n'
          'Settings: {}\n'
          'Override: {}\n'
          'Init memory: {:.1f} GiB\n'
          'Peak memory: {:.1f} GiB\n'
          'After memory: {:.1f} GiB\n'
          'Estimated model memory: {:.1f} GiB\n'
          'Times for each step: {}\n'
          'Time for encoding: {:.4f} s\n'
          'Time for decoding 10 tokens: {:.4f} s +/- {} %\n'
          '\n\n'
          .format(size, settings, override_settings,
                  init_memory/1024**3, peak_memory/1024**3,
                  after_memory/1024**3, model_memory_gigabytes,
                  elapsed_times, encoding_time,
                  decoding_time_10, decoding_time_diff_percent))
      print(message)
      messages.append(message)
      encoding_times.append(encoding_time)
      decoding_times.append(decoding_time_10)
      sizes.append(size)
      memories.append(model_memory_gigabytes)

    print('Final results (recap):')
    for message in messages:
      print(message)

    # This is useful for copying results into a spreadsheet etc.
    # for i in range(len(sweep_settings)):
    #   print('{}\t{}\t{}\t{:.1f}'.format(
    #       sizes[i], encoding_times[i], decoding_times[i], memories[i]))

  def test_loss_layer_timing(self):
    all_settings = [
        # The first run is sometimes slower, less reliable.
        {'output': 32000, 'input': 2048, 'prob': None,
         'type': None, 'sparsity': 0, 'lowrank': 0, 'use_bias': False},

        {'output': 32000, 'input': 2048, 'prob': None,
         'type': None, 'sparsity': 0, 'lowrank': 0, 'use_bias': False},
        {'output': 32000, 'input': 2048, 'prob': None,
         'type': 'einsum', 'sparsity': 0, 'lowrank': 0, 'use_bias': False},
        {'output': 32000, 'input': 2048, 'prob': None,
         'type': 'mult', 'sparsity': 2, 'lowrank': 0, 'use_bias': False},

        {'output': 32000, 'input': 2048, 'prob': None,
         'type': None, 'sparsity': 0, 'lowrank': 0, 'use_bias': True},
        {'output': 32000, 'input': 2048, 'prob': None,
         'type': 'einsum', 'sparsity': 0, 'lowrank': 0, 'use_bias': True},
        {'output': 32000, 'input': 2048, 'prob': None,
         'type': 'mult', 'sparsity': 2, 'lowrank': 0, 'use_bias': True},
    ]

    messages = []
    for settings in all_settings:
      pred_model = tl.SparseDenseWithOptions(
          n_units=settings['output'],
          d_input=settings['input'],
          sparsity_type=settings['type'],
          sparsity=settings['sparsity'],
          d_lowrank=settings['lowrank'],
          prob_sparse=settings['prob'],
          use_bias=settings['use_bias'],
          mode='predict',
          )
      pred_model = tl.Accelerate(pred_model)

      shape1l = shapes.ShapeDtype((1, settings['input']))
      pred_model.init(input_signature=shape1l)
      inputs = np.ones((1, settings['input']))

      total_time = 0.0
      for counter in range(-50, 100):
        start_time = time.time()
        y = pred_model(inputs)
        self.assertEqual(y.shape, (1, settings['output']))
        elapsed_time = time.time() - start_time
        if counter >= 0:
          total_time += elapsed_time

      message = (
          '\n\nParams: %d Settings: %s\nTime for 100 tokens: %.4f s\n\n\n'
          % (_size_of_model(pred_model), settings, total_time))
      messages.append(message)
      print(message)

    print('Final results (recap):')
    for message in messages:
      print(message)


if __name__ == '__main__':
  config.config_with_absl()
  test.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for accelerated optimization of loss layers."""

import time
from absl.testing import absltest

from jax import test_util  # pylint: disable=unused-import
from jax.config import config
import numpy as np

from trax import fastmath
from trax import layers as tl
from trax import optimizers
from trax import shapes
from trax.layers import base
from trax.models.research import terraformer


class TrainerTest(absltest.TestCase):

  def _assert_all_equal(self, t1, t2, tol=1e-5):
    def eq(x1, x2):
      diff = np.maximum(np.abs(x1 - x2) - tol, 0.0)
      self.assertLessEqual(np.sum(diff), 0.0,
                           msg=f'\n{x1}\n !=\n{x2}\n diff:\n{x1-x2}')
    fastmath.nested_map_multiarg(eq, t1, t2)

  def test_run_simple_task(self):
    """Runs an accelerated optimizer on a simple task."""
    inputs_batch = np.arange(8).reshape((8, 1))  # 8 items per batch
    targets_batch = np.pi * np.ones_like(inputs_batch)
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    loss_layer = tl.Serial(tl.Dense(1), tl.L2Loss())
    loss_layer.init(labeled_batch)
    optimizer = optimizers.SGD(.01)
    optimizer.tree_init(loss_layer.weights)
    trainer = optimizers.Trainer(loss_layer, optimizer)
    rng = fastmath.random.get_prng(0)
    trainer.one_step(labeled_batch, rng)


  def test_run_sharded_terraformer(self):
    """Runs Terraformer with sharded weights (only on 2+-device systems)."""
    if fastmath.local_device_count() == 1:
      return
    base.N_WEIGHTS_SHARDS = fastmath.local_device_count()
    inputs_batch = np.arange(8).reshape((2, 4)) + 1
    targets_batch = 2 * inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    int_sig = shapes.ShapeDtype((2, 4), dtype=np.int32)
    input_sig = (int_sig, int_sig, int_sig)
    # We want to test rng propagation too, so adding some dropout layers.
    model = terraformer.ConfigurableTerraformer(
        20, d_model=8, d_ff=32, n_heads=1, dropout=0.0,
        n_encoder_layers=2, n_decoder_layers=2,
        ff_sparsity=(4, 8, 0.0, 1.0),
        encoder_attention_type=tl.Attention,
        encoder_decoder_attention_type=tl.CausalAttention,
        pos_type=None, reversible_encoder=True)
    loss = tl.Serial(tl.LogSoftmax(), tl.CrossEntropyLoss())
    model_with_loss = tl.Serial(model, loss)
    rng_init = fastmath.random.get_prng(12)
    model_with_loss.init(input_sig, rng=rng_init)

    # Make a step with the trainer.
    optimizer = optimizers.Adafactor(0.01)
    split_w = fastmath.nested_map(
        lambda x: x[0],
        tl.shard(model_with_loss.weights, base.N_WEIGHTS_SHARDS))
    optimizer.tree_init(split_w)
    trainer = optimizers.Trainer(model_with_loss, optimizer)
    rng_step1 = fastmath.random.get_prng(7)
    trainer.one_step(labeled_batch, rng_step1)
    # Reset shards back to default.
    base.N_WEIGHTS_SHARDS = 1

  def test_run_reversible_slots(self):
    """Tests that slots can be read and assigned in reversible trainer."""
    layers = [tl.Dense(4), tl.Dup()]
    rev_layers = [tl.ReversibleHalfResidual(tl.Dense(4)),
                  tl.ReversibleSwap()]
    loss_layer = tl.Serial(tl.Concatenate(), tl.Dense(4),
                           tl.LogSoftmax(), tl.CrossEntropyLoss())
    trainer = optimizers.ReversibleSerialTrainer(
        [(layers, rev_layers)], loss_layer, optimizers.Adam)
    slots = trainer.slots
    trainer.slots = slots
    self.assertEqual(slots, trainer.slots)

  def test_run_reversible_same_as_default_basic(self):
    """Runs the reversible trainer, check results are the same as default."""
    inputs_batch = np.arange(8).reshape((2, 4))
    targets_batch = 2 * inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    # We want to test rng propagation too, so adding some dropout layers.
    first_layer = tl.Serial(tl.Embedding(9, 4), tl.Dropout(0.5), tl.Dup())
    rev_layers = [tl.ReversibleHalfResidual(tl.Dense(4), tl.Dropout(0.2)),
                  tl.ReversibleSwap(),
                  tl.ReversibleHalfResidual(tl.Dropout(0.5), tl.Dense(4)),
                  tl.ReversibleSwap()]
    loss_layer = tl.Serial(tl.Concatenate(), tl.Dense(19), tl.Dropout(0.3),
                           tl.LogSoftmax(), tl.CrossEntropyLoss())
    model = tl.Serial([first_layer] + rev_layers + [loss_layer])
    rng_init = fastmath.random.get_prng(12)
    model.init(labeled_batch, rng=rng_init)
    optimizer_fn = optimizers.Adam  # to test slots

    # Make 2 steps with the original trainer.
    optimizer = optimizer_fn()
    optimizer.tree_init(model.weights)
    trainer = optimizers.Trainer(model, optimizer)
    rng_step1 = fastmath.random.get_prng(7)
    rng_step2 = fastmath.random.get_prng(8)
    trainer.one_step(labeled_batch, rng_step1)
    trainer.one_step(labeled_batch, rng_step2, learning_rate=0.02)
    first_layer_weights1 = first_layer.weights
    rev_layer0_weights1 = rev_layers[0].weights
    rev_layer2_weights1 = rev_layers[2].weights
    loss_layer_weights1 = loss_layer.weights

    # Now make 2 steps with reversible trainer.
    model.init(labeled_batch, rng=rng_init)
    trainer = optimizers.ReversibleSerialTrainer(
        [(first_layer.sublayers, rev_layers)], loss_layer, optimizer_fn)
    trainer.one_step(labeled_batch, rng_step1)
    trainer.one_step(labeled_batch, rng_step2, learning_rate=0.02)

    # Check that weights end up the same.
    self._assert_all_equal(loss_layer_weights1, loss_layer.weights)
    self._assert_all_equal(rev_layer2_weights1, rev_layers[2].weights)
    self._assert_all_equal(rev_layer0_weights1, rev_layers[0].weights)
    self._assert_all_equal(first_layer_weights1, first_layer.weights)

  def test_run_reversible_same_as_default_extended(self):
    """Runs the reversible trainer, check results are the same as default."""
    inputs_batch = np.arange(8).reshape((2, 4))
    targets_batch = 2 * inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    # We want to test rng propagation too, so adding some dropout layers.
    first_layer = tl.Serial(tl.Embedding(9, 4), tl.Dropout(0.5), tl.Dup())
    rev_layers1 = [tl.ReversibleHalfResidual(tl.Dense(4), tl.Dropout(0.2)),
                   tl.ReversibleSwap(),
                   tl.ReversibleHalfResidual(tl.Dropout(0.5), tl.Dense(4)),
                   tl.ReversibleSwap()]
    mid_layer = tl.Serial(tl.Add(), tl.Dense(4), tl.Dup())
    rev_layers2 = [tl.ReversibleHalfResidual(tl.Dense(4), tl.Dropout(0.3)),
                   tl.ReversibleSwap()]
    loss_layer = tl.Serial(tl.Concatenate(), tl.Dense(19), tl.Dropout(0.3),
                           tl.LogSoftmax(), tl.CrossEntropyLoss())
    model = tl.Serial([first_layer] + rev_layers1 + [mid_layer] +
                      rev_layers2 + [loss_layer])
    rng_init = fastmath.random.get_prng(12)
    model.init(labeled_batch, rng=rng_init)
    optimizer_fn = optimizers.Adam  # to test slots

    # Make 3 steps with the original trainer.
    optimizer = optimizer_fn()
    optimizer.tree_init(model.weights)
    trainer = optimizers.Trainer(model, optimizer)
    rng_step1 = fastmath.random.get_prng(7)
    rng_step2 = fastmath.random.get_prng(8)
    rng_step3 = fastmath.random.get_prng(9)
    trainer.one_step(labeled_batch, rng_step1)
    trainer.one_step(labeled_batch, rng_step2, learning_rate=0.02)
    trainer.one_step(labeled_batch, rng_step3, learning_rate=0.03)
    first_layer_weights1 = first_layer.weights
    rev_layer12_weights1 = rev_layers1[2].weights
    mid_layer_weights1 = mid_layer.weights
    rev_layer20_weights1 = rev_layers2[0].weights
    loss_layer_weights1 = loss_layer.weights

    # Now make 3 steps with reversible trainer.
    model.init(labeled_batch, rng=rng_init)
    # TODO(lukaszkaiser): this test seems to fail with memoize_jit, why?
    trainer = optimizers.ReversibleSerialTrainer(
        [(first_layer.sublayers, rev_layers1),
         (mid_layer.sublayers, rev_layers2)],
        loss_layer, optimizer_fn, memoize_jit=False)
    trainer.one_step(labeled_batch, rng_step1)
    trainer.one_step(labeled_batch, rng_step2, learning_rate=0.02)
    trainer.one_step(labeled_batch, rng_step3, learning_rate=0.03)

    # Check that weights end up the same.
    self._assert_all_equal(loss_layer_weights1, loss_layer.weights)
    self._assert_all_equal(rev_layer20_weights1, rev_layers2[0].weights)
    self._assert_all_equal(mid_layer_weights1, mid_layer.weights)
    self._assert_all_equal(rev_layer12_weights1, rev_layers1[2].weights)
    self._assert_all_equal(first_layer_weights1, first_layer.weights)

  def test_run_reversible_same_as_default_terraformer(self):
    """Runs the reversible trainer, check results are the same as default."""
    inputs_batch = np.arange(8).reshape((2, 4)) + 1
    targets_batch = 2 * inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    int_sig = shapes.ShapeDtype((2, 4), dtype=np.int32)
    input_sig = (int_sig, int_sig, int_sig)
    # We want to test rng propagation too, so adding some dropout layers.
    model = terraformer.ConfigurableTerraformer(
        20, d_model=8, d_ff=32, n_heads=1, dropout=0.0, n_encoder_layers=2,
        n_decoder_layers=2, ff_sparsity=(4, 8, 0.0, 1.0), pos_type=None,
        reversible_encoder=True)
    loss = tl.Serial(tl.LogSoftmax(), tl.CrossEntropyLoss())
    optimizer_fn = optimizers.Adafactor
    blocks, loss_layer = optimizers.trainer.extract_reversible_blocks(
        [model, loss], loss_chunk_size=4)
    blocks_serial = [(tl.Serial(std), rev) for (std, rev) in blocks]
    model_with_loss = tl.Serial(model, loss)
    rng_init = fastmath.random.get_prng(12)
    model_with_loss.init(input_sig, rng=rng_init)

    # Make 3 steps with the original trainer.
    optimizer = optimizer_fn()
    optimizer.tree_init(model_with_loss.weights)
    trainer = optimizers.Trainer(model_with_loss, optimizer)
    rng_step1 = fastmath.random.get_prng(7)
    rng_step2 = fastmath.random.get_prng(8)
    rng_step3 = fastmath.random.get_prng(9)
    trainer.one_step(labeled_batch, rng_step1)
    trainer.one_step(labeled_batch, rng_step2, learning_rate=0.02)
    trainer.one_step(labeled_batch, rng_step3, learning_rate=0.03)
    first_weights = blocks_serial[0][0].weights
    first_rev_weights = blocks[0][1][0].weights
    loss_weights = loss_layer.weights

    # Now make 3 steps with reversible trainer.
    model_with_loss.init(input_sig, rng=rng_init)
    trainer = optimizers.ReversibleSerialTrainer(
        blocks, loss_layer, optimizer_fn)
    trainer.one_step(labeled_batch, rng_step1)
    trainer.one_step(labeled_batch, rng_step2, learning_rate=0.02)
    trainer.one_step(labeled_batch, rng_step3, learning_rate=0.03)

    # Check that weights end up the same.
    self._assert_all_equal(loss_weights, loss_layer.weights)
    self._assert_all_equal(first_rev_weights, blocks[0][1][0].weights)
    self._assert_all_equal(first_weights, blocks_serial[0][0].weights)

  def test_run_reversible_large_weights(self):
    """Runs the reversible trainer with a lot of weights to test memory use."""
    # This test requires > 18GB RAM, only run on TPUs. It does pass on GPU
    # and CPU when you run it locally, but it's too big for unit-testing.
    ram_limited = True  # Set to False to run this test locally.
    if fastmath.global_device_count() == 1 and ram_limited:
      return

    # Create inputs and rngs.
    inputs_batch = np.arange(8).reshape((2, 4))
    targets_batch = inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    first_layer = tl.Serial(tl.Embedding(9, 16*1024), tl.Dup())
    rng_init = fastmath.random.get_prng(12)
    rng_step = fastmath.random.get_prng(13)

    # Initialize layers.
    first_layer.init(labeled_batch, rng=rng_init)
    n_layers = 18  # 18 layers each 16K x 16K = 256M weights ~= 1GB, 18GB ram
    rev_layers = []
    int_shape = shapes.ShapeDtype((2, 4), dtype=np.int32)
    shape = shapes.ShapeDtype((2, 4, 16*1024))
    sig = (shape, shape)
    for _ in range(n_layers):
      layer = tl.ReversibleHalfResidual(tl.Dense(16*1024))
      layer.init(sig, rng=rng_init)
      layer.weights = tl.on_cpu(layer.weights)  # store weights in cpu memory
      rev_layers.append(layer)
      rev_layers.append(tl.ReversibleSwap())
    loss_layer = tl.Serial(tl.Concatenate(), tl.Dense(9),
                           tl.LogSoftmax(), tl.CrossEntropyLoss())
    loss_layer.init((shape, shape, int_shape, int_shape))
    optimizer_fn = optimizers.Adafactor

    # Make a step with reversible trainer.
    trainer = optimizers.ReversibleSerialTrainer(
        [(first_layer, rev_layers)], loss_layer, optimizer_fn)
    loss, _ = trainer.one_step(labeled_batch, rng_step)
    self.assertLess(float(loss.sum()), 10000.0)  # Just to get the loss.
    # Set to true to run again, e.g., for profiling.
    run_twice = False
    if run_twice:
      t = time.time()
      loss, _ = trainer.one_step(labeled_batch, rng_step)
      self.assertLess(float(loss.sum()), 10000.0)  # Just to get the loss.
      print('Took %.3f seconds to run, loss %s' % (time.time() - t, loss))

  def test_run_reversible_weights_trainsfer_xprof(self):
    """Runs the reversible trainer and profiles weight transfer stats."""
    run_this_test = False  # We only run this test manually.
    if not run_this_test or fastmath.global_device_count() == 1:  # TPU only
      return

    # Create inputs and rngs.
    inputs_batch = np.ones((1024, 128), dtype=np.int32)
    targets_batch = inputs_batch
    labeled_batch = (inputs_batch, targets_batch, np.ones_like(targets_batch))
    first_layer = tl.Serial(tl.Embedding(4, 1024), tl.Dup())
    rng_init = fastmath.random.get_prng(12)
    rng_step = fastmath.random.get_prng(13)

    # Initialize layers.
    first_layer.init(labeled_batch, rng=rng_init)
    n_layers = 6
    rev_layers = []
    int_shape = shapes.ShapeDtype((1024, 128), dtype=np.int32)
    shape = shapes.ShapeDtype((1024, 128, 1024))
    sig = (shape, shape)
    for _ in range(n_layers):
      layer = tl.ReversibleHalfResidual(tl.Dense(1024))
      layer.init(sig, rng=rng_init)
      layer.weights = tl.on_cpu(layer.weights)  # store weights in cpu memory
      rev_layers.append(layer)
      rev_layers.append(tl.ReversibleSwap())
    loss_layer = tl.Serial(tl.Concatenate(), tl.Dense(9),
                           tl.LogSoftmax(), tl.CrossEntropyLoss())
    loss_layer.init((shape, shape, int_shape, int_shape))
    optimizer_fn = optimizers.SGD

    # Make a step with reversible trainer.
    trainer = optimizers.ReversibleSerialTrainer(
        [(first_layer, rev_layers)], loss_layer, optimizer_fn)
    loss, _ = trainer.one_step(labeled_batch, rng_step)
    self.assertLess(float(loss.sum()), 10000.0)  # Just to get the loss.
    # We profile here.
    t = time.time()
    loss, _ = trainer.one_step(labeled_batch, rng_step)
    self.assertLess(float(loss.sum()), 10000.0)  # Just to get the loss.
    print('Took %.3f seconds to run, loss %s' % (time.time() - t, loss))


if __name__ == '__main__':
  config.config_with_absl()
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dtypes and dtype utilities."""
import numpy as np

# We use numpy's dtypes instead of TF's, because the user expects to use them
# with numpy facilities such as `np.dtype(np.int64)` and
# `if x.dtype.type is np.int64`.
# pylint: disable=unused-import
# pylint: disable=g-bad-import-order
from numpy import bool_
from numpy import int_
from numpy import int16
from numpy import int32
from numpy import int64
from numpy import int8
from numpy import uint16
from numpy import uint32
from numpy import uint64
from numpy import uint8
from numpy import float_
from numpy import float16
from numpy import float32
from numpy import float64
from numpy import complex_
from numpy import complex64
from numpy import complex128

from numpy import inexact

from numpy import iinfo
from numpy import issubdtype

from numpy import inf

# TODO(wangpeng): Make bfloat16 a numpy dtype instead of using TF's
from tensorflow.compat.v2 import bfloat16
# pylint: enable=g-bad-import-order
# pylint: enable=unused-import


_to_float32 = {
    np.dtype('float64'): np.dtype('float32'),
    np.dtype('complex128'): np.dtype('complex64'),
}


_allow_float64 = True


def is_allow_float64():
  return _allow_float64


def set_allow_float64(b):
  global _allow_float64
  _allow_float64 = b


def canonicalize_dtype(dtype):
  if not is_allow_float64():
    return _to_float32.get(dtype, dtype)
  else:
    return dtype


def _result_type(*arrays_and_dtypes):
  dtype = np.result_type(*arrays_and_dtypes)
  return canonicalize_dtype(dtype)


def default_float_type():
  """Gets the default float type.

  Returns:
    If `is_allow_float64()` is true, returns float64; otherwise returns float32.
  """
  if is_allow_float64():
    return float64
  else:
    return float32



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Combinators for composing layers."""

import copy

from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import base
from trax.layers.base import Fn
from trax.shapes import ShapeDtype


class Serial(base.Layer):
  """Combinator that applies layers serially (by function composition).

  This combinator is commonly used to construct deep networks, e.g., like this::

      mlp = tl.Serial(
        tl.Dense(128),
        tl.Relu(),
        tl.Dense(10),
      )

  A Serial combinator uses stack semantics to manage data for its sublayers.
  Each sublayer sees only the inputs it needs and returns only the outputs it
  has generated. The sublayers interact via the data stack. For instance, a
  sublayer k, following sublayer j, gets called with the data stack in the
  state left after layer j has applied. The Serial combinator then:

    - takes n_in items off the top of the stack (n_in = k.n_in) and calls
      layer k, passing those items as arguments; and

    - takes layer k's n_out return values (n_out = k.n_out) and pushes
      them onto the data stack.

  A Serial instance with no sublayers acts as a special-case (but useful)
  1-input 1-output no-op.
  """

  def __init__(self, *sublayers, name=None, sublayers_to_print=None):
    super().__init__(
        name=name, sublayers_to_print=sublayers_to_print)

    sublayers = _ensure_flat(sublayers)
    self._sublayers = sublayers
    self._n_layers = len(sublayers)

    if sublayers:
      self._n_in, self._n_out = self._n_inputs_n_outputs(sublayers)
      self._weights = tuple(None for l in sublayers)
      self._state = tuple(None for l in sublayers)

  def forward(self, xs):
    """Executes this layer as part of a forward pass through the model."""
    self._validate_forward_inputs(xs)
    if not self.sublayers:  # No-op: outputs = inputs
      return xs

    state, weights = self.state, self.weights
    rngs = _split_rngs(self.rng, self._n_layers)
    stack = xs
    new_state = []
    n_layers = self._n_layers
    if len(weights) != n_layers:
      raise ValueError(
          f'Number of weight elements ({len(weights)}) does not equal '
          f'number of sublayers ({n_layers}).')
    if len(state) != n_layers:
      raise ValueError(
          f'Number of state elements ({len(state)}) does not equal '
          f'number of sublayers ({n_layers}).')

    for layer, w, s, rng in zip(self.sublayers, weights, state, rngs):
      inputs = inputs_from_stack(stack, layer.n_in)
      outputs, s = layer.pure_fn(inputs, w, s, rng, use_cache=True)
      stack = outputs_onto_stack(outputs, stack, layer.n_in)
      new_state.append(s)
    self.state = tuple(new_state)
    return stack

  # pylint: disable=protected-access
  def init_weights_and_state(self, input_signature):
    """Initializes weights and state for inputs with the given signature."""
    weights = []
    states = []
    # In the code below, stack, inputs, and outputs are abstract (shapes and
    # dtypes), but weights and states are non-abstract actual values.
    stack = input_signature
    for sublayer in self.sublayers:
      inputs = inputs_from_stack(stack, sublayer.n_in)
      weights_or_cache_marker, state_or_cache_marker = (
          sublayer.init(inputs, use_cache=True))
      outputs, _ = sublayer._forward_abstract(inputs)
      stack = outputs_onto_stack(outputs, stack, sublayer.n_in)

      weights.append(weights_or_cache_marker)
      states.append(state_or_cache_marker)
    self.state = tuple(states)
    self.weights = tuple(weights)
  # pylint: enable=protected-access

  def _n_inputs_n_outputs(self, layers):
    del self
    running_max = 0
    running_total = 0
    for layer in layers:
      running_total += layer.n_in
      running_max = max(running_max, running_total)
      running_total -= layer.n_out
    return running_max, (running_max - running_total)

  def _validate_forward_inputs(self, xs):
    if not isinstance(xs, (tuple, list)) and self._n_in != 1:
      raise TypeError(f'Serial.forward input must be a tuple or list; '
                      f'instead got {type(xs)}.')
      # TODO(jonni): Include full xs (or shape) in error message?
    len_xs = 1 if isinstance(xs, jnp.ndarray) else len(xs)
    if len_xs < self.n_in:
      raise ValueError(
          f'Number of inputs ({len(xs)}) to Serial.forward less than n_in '
          f'({self.n_in}).')


class Parallel(base.Layer):
  """Combinator that applies a list of layers in parallel to its inputs.

  Layers in the list apply to successive spans of inputs, where the spans are
  determined how many inputs each layer takes. The resulting output is the
  (flattened) concatenation of the respective layer outputs.

  For example, suppose one has three layers:

    - F: 1 input, 1 output
    - G: 3 inputs, 1 output
    - H: 2 inputs, 2 outputs (h1, h2)

  Then Parallel(F, G, H) will take 6 inputs and give 4 outputs:

    - inputs: a, b, c, d, e, f
    - outputs: F(a), G(b, c, d), h1, h2     where h1, h2 = H(e, f)

  As an important special case, a None argument to Parallel acts as if it takes
  one argument, which it leaves unchanged. (It acts as a one-arg no-op.) For
  example:

    Parallel(None, F)

  creates a layer that passes its first input unchanged and applies F to the
  following input(s).
  """

  def __init__(self, *sublayers, name=None):
    """The constructor.

    Args:
      *sublayers: A list of sublayers.
      name: Descriptive name for this layer.

    Returns:
      A new layer in which each of the given sublayers applies to its
      corresponding span of elements in the dataflow stack.
    """
    super().__init__(name=name)
    sublayers = self._validate(sublayers)
    self._n_layers = len(sublayers)
    self._sublayers = sublayers
    self._n_in = sum(l.n_in for l in sublayers)
    self._n_out = sum(l.n_out for l in sublayers)
    self._weights = tuple(None for l in sublayers)
    self._state = tuple(None for l in sublayers)

  def forward(self, inputs):
    """Executes this layer as part of a forward pass through the model."""
    n_layers, layers = self._n_layers, self.sublayers
    sublayer_inputs = self._allot_to_sublayers(inputs)
    state, weights = self.state, self.weights
    rngs = _split_rngs(self.rng, n_layers)
    if len(sublayer_inputs) != n_layers:
      raise ValueError(
          f'Number of inputs for sublayers ({len(sublayer_inputs)}) does not equal '
          f'number of sublayers ({n_layers}).')
    if len(weights) != n_layers:
      raise ValueError(
          f'Number of weight elements ({len(weights)}) does not equal '
          f'number of sublayers ({n_layers}).')
    if len(state) != n_layers:
      raise ValueError(
          f'Number of state elements ({len(state)}) does not equal '
          f'number of sublayers ({n_layers}).')
    if len(rngs) != n_layers:
      raise ValueError(
          f'Number of rngs ({len(rngs)}) does not equal '
          f'number of sublayers ({n_layers}).')
    outputs = []
    new_state = []
    for layer, x, w, s, r in zip(layers, sublayer_inputs, weights, state, rngs):
      # Note that zip silently truncates its result if lengths don't match.
      sub_outputs, sub_state = layer.pure_fn(x, w, s, r, use_cache=True)
      if layer.n_out == 1:
        outputs.append(sub_outputs)
      else:
        outputs.extend(sub_outputs)
      new_state.append(sub_state)
    output = outputs[0] if self.n_out == 1 else tuple(outputs)
    self.state = tuple(new_state)
    return output

  def init_weights_and_state(self, input_signature):
    """Initializes weights and state for inputs with the given signature."""
    sublayer_signatures = self._allot_to_sublayers(input_signature)
    inits = [layer.init(signature, use_cache=True)
             for layer, signature
             in zip(self.sublayers, sublayer_signatures)]
    if inits:
      weights, state = tuple(zip(*inits))
      self.state = state
      self.weights = weights

  def _validate(self, layers):
    if not layers or len(layers) < 2:
      raise ValueError(
          f'layers ({layers}) must be a list with at least two elements')
    layers = list(layers)  # Ensure we can modify layers.
    for i, obj in enumerate(layers):
      if obj is None or obj == []:  # pylint: disable=g-explicit-bool-comparison
        layers[i] = Serial(None)
      elif isinstance(obj, (list, tuple)):
        layers[i] = Serial(obj)
      else:
        if not isinstance(obj, base.Layer):
          raise ValueError(
              f'Found nonlayer object ({obj}) in layers list: [{layers}]')
      if layers[i].n_in == 0:
        raise ValueError(
            f'Sublayer with n_in = 0 not allowed in Parallel: {layers[i]}')
    return layers

  def _allot_to_sublayers(self, inputs):
    """Divides Parallel's inputs for use by the sublayers.

    Args:
      inputs: Tuple of ndarrays or ShapeDtype instances.

    Returns:
      A tuple that partitions this layer's inputs among its sublayers.
      Sublayers that take one argument get that argument directly. All other
      sublayers get a tuple of items.
    """
    start, end = 0, 0
    sub_inputs = []
    for layer in self.sublayers:
      n_in = layer.n_in
      end = start + n_in
      if n_in == 1:
        sub_inputs.append(inputs[start])
      else:
        sub_inputs.append(inputs[start:end])
      start = end
    return tuple(sub_inputs)


class Concatenate(base.Layer):
  """Concatenates a number of tensors into a single tensor.

  For example::

      x = np.array([1, 2])
      y = np.array([3, 4])
      z = np.array([5, 6])
      concat3 = tl.Concatenate(n_items=3)
      z = concat3((x, y, z))  # z = [1, 2, 3, 4, 5, 6]

  Use the `axis` argument to specify on which axis to concatenate the tensors.
  By default it's the last axis, `axis=-1`, and `n_items=2`.
  """

  def __init__(self, n_items=2, axis=-1):
    name = 'Concatenate' if axis == -1 else f'Concatenate_axis{axis}'
    super().__init__(n_in=n_items, name=name)
    self._n_items = n_items
    self._axis = axis

  def forward(self, xs):
    """Executes this layer as part of a forward pass through the model."""
    return jnp.concatenate(xs, self._axis)


class Split(base.Layer):
  """Splits the input into n items along an axis."""

  def __init__(self, n_items=2, axis=-1):
    super().__init__(n_out=n_items)
    self._n_items = n_items
    self._axis = axis

  def forward(self, inputs):
    """Executes this layer as part of a forward pass through the model."""
    return tuple(jnp.split(inputs, self._n_items, self._axis))


def _scan(f, xs, init_value, axis=0, remat=False):
  """Scans the f over the given axis of xs.

  In pseudo-python, the scan function would look as follows:

  def scan(f, xs, init_value, axis):
    xs  = [xs[..., i, ...] for i in range(xs.shape[axis])]
    cur_value = init_value
    ys = []
    for x in xs:
      y, cur_value = f(x, cur_value)
      ys.append(y)
    return np.stack(ys, axis), cur_value

  Args:
    f: function (x, carry) -> (y, new_carry)
    xs: tensor, x will be xs slices on axis
    init_value: tensor, initial value of the carry-over
    axis: int, the axis on which to slice xs
    remat: whether to re-materialize f

  Returns:
    A pair (ys, last_value) as described above.
  """
  def swapaxes(x):
    transposed_axes = list(range(len(x.shape)))
    transposed_axes[axis] = 0
    transposed_axes[0] = axis
    return jnp.transpose(x, axes=transposed_axes)
  if axis != 0:
    xs = fastmath.nested_map(swapaxes, xs)
  def transposed_f(c, x):
    y, d = f(x, c)
    return d, y
  if remat:
    transposed_f = fastmath.remat(transposed_f)
  last_value, ys = fastmath.scan(transposed_f, init_value, xs)
  if axis != 0:
    ys = fastmath.nested_map(swapaxes, ys)
  return ys, last_value


class Scan(base.Layer):
  """Applies a layer progressively/cumulatively to an axis-derived sequence.

  Conceptually, this is a function from a list to a same-length list of partial
  (cumulative) results. For instance, a list of values (`[1, 2, 3, 4, 5]`) can
  transform to a list of cumulative sums (`[1, 3, 6, 10, 15]`). Functions for
  the same concept are called `scan` in Scala, `scanl` in Haskell, and
  `accumulate*` in Factor.

  In more detail, we assume the layer takes a tuple of inputs of the following
  form:

    (input1, ..., inputN, carry1, ..., carryM)

  and returns:

    (output1, ..., outputK, new_carry1, ..., new_carryM)

  The scanned version applies the layer iteratively to a tensor treating values
  at the given axis as if they were a list. For example, to calculate all
  sums of prefixes of a tensor, we can do this::

    def add(x, carry):
      def f(input, carry):
        res = input + carry
        return res, res  # output and carry are the same
      return tl.Fn('add', f, n_out=2)

    Scan(add)([1, 2, 3], 0) = [1, 3, 6], 6
  """

  def __init__(self, layer, axis=0, n_carry=1, remat=False, mode='train'):
    super().__init__(n_in=layer.n_in, n_out=layer.n_out)
    self._sublayers = [layer]
    self._n_carry = n_carry
    self._axis = axis
    self._remat = remat
    self._weights = (None,)
    self._state = (None, ())
    self._mode = mode

  @property
  def sublayer(self):
    """Returns the unique sublayer managed by this layer."""
    return self._sublayers[0]

  @property
  def state(self):
    """Returns a tuple containing this layer's state."""
    return (self.sublayer.state, self._state[1])

  @state.setter
  def state(self, state):
    """Recursively sets state on this layer the sublayer."""
    if isinstance(state, dict) and state == base.GET_STATE_FROM_CACHE:
      return
    self._state = (None, state[1])
    self.sublayer.state = state[0]

  def forward(self, inputs):
    """Executes this layer as part of a forward pass through the model."""
    weights = self.weights[0]
    if isinstance(inputs, list):
      inputs = tuple(inputs)  # so that inputs structure matches outputs
    n_carry = self._n_carry
    def scannable_fn(x, carry_and_state):  # pylint: disable=invalid-name
      carry, state, i = carry_and_state
      x_and_carry = x + carry if n_carry > 0 else x
      rng = fastmath.random.fold_in(self.rng, i)
      res, new_state = self.sublayer.pure_fn(
          x_and_carry, weights, state, rng, use_cache=True)
      if n_carry > 0:
        return (res[:-n_carry], (res[-n_carry:], new_state, i+1))
      else:
        return (res, ([], new_state, i+1))

    if n_carry > 0:
      xs = inputs[:-n_carry]  # Split input stack into inputs and carry.
      xs_carry = inputs[-n_carry:]
      if self._mode == 'predict' and self._state[1] is not ():  # pylint: disable=literal-comparison
        xs_carry = self._state[1]
      init = (xs_carry, self.state[0], jnp.array(0, dtype=jnp.int32))
    else:
      xs_carry = ()
      xs, init = inputs, ([], self.state[0], jnp.array(0, dtype=jnp.int32))
    ys, (carry, new_state, _) = _scan(scannable_fn, xs, init,
                                      axis=self._axis, remat=self._remat)
    res = ys + carry if n_carry > 0 else ys
    state_carry = carry if self._mode == 'predict' and n_carry > 0 else ()
    self.state = (new_state, state_carry)
    return res  # Put outputs and carry back on stack.

  def init_weights_and_state(self, input_signature):
    """Initializes weights and state for inputs with the given signature."""
    n_carry = self._n_carry
    if n_carry == 0:
      if isinstance(input_signature, (list, tuple)):
        layer_sig = [ShapeDtype(_shape_without_axis(x, self._axis), x.dtype)
                     for x in input_signature]
        layer_sig = tuple(layer_sig)
      else:
        layer_sig = ShapeDtype(_shape_without_axis(input_signature, self._axis),
                               input_signature.dtype)
      weights, state = self.sublayer.init(layer_sig)
      self.state = (state, ())
      self.weights = (weights,)
    else:
      xs = input_signature[:-n_carry]
      init = input_signature[-n_carry:]
      xs_slices = [ShapeDtype(_shape_without_axis(x, self._axis), x.dtype)
                   for x in xs]
      layer_signature = tuple(xs_slices + list(init))
      weights, state = self.sublayer.init(layer_signature, use_cache=True)
      self.state = (state, ())
      self.weights = (weights,)


class Cond(base.Layer):
  """Applies layers conditionally.

  For parameters `cond`, `true`, and `false` runs the equivalent of `true(y)
  if cond(x) else false(y)`, where `x` is `cond.n_in` elements from front of the
  stack and `y` is the rest of the stack.
  Exactly one of `true` and `false` functions is executed, so it can be used to
  conditionally run long computations. The state of non-executed function is not
  updated. Note that different branches may be executed on different devices
  if `cond` returns different values on them.
  By default 'false' function is an identity.

  `cond` must return exactly one element: a Boolean value.
  `true` and `false` must have the same n_in, and the same n_out.
  """

  def __init__(self, cond, true, false=None, name=None):
    super(Cond, self).__init__(name=name)

    if false is None:
      self._identity_false_fun = True
      # We don't need this function, but it will be useful for checking if
      # 'true' has proper n_in/n_out.
      false = Serial()
      self._false = false
    else:
      self._identity_false_fun = False
      self._false = false

    sublayers = [cond, true, false]
    self._sublayers = sublayers
    self._n_layers = len(sublayers)
    self._cond = cond
    self._true = true

    if cond.n_out != 1:
      raise ValueError(
          'cond.n_out must be 1: cond:{}->{}'.format(cond.n_in, cond.n_out))
    if true.n_in != false.n_in:
      raise ValueError(
          'true.n_in and false.n_in must be equal: true:{}->{} ; false:{}->{}'
          .format(true.n_in, true.n_out, false.n_in, false.n_out))
    if true.n_out != false.n_out:
      raise ValueError(
          'true.n_out and false.n_out must be equal: true:{}->{} ; false:{}->{}'
          .format(true.n_in, true.n_out, false.n_in, false.n_out))

    self._n_in = cond.n_in + true.n_in
    self._n_out = true.n_out
    self._weights = tuple(None for l in sublayers)
    self._state = tuple(None for l in sublayers)

  # pylint: disable=protected-access
  def init_weights_and_state(self, input_signature):
    """Initializes weights and state for inputs with the given signature."""
    weights = []
    states = []
    # In the code below, stack, inputs, and outputs are abstract (shapes and
    # dtypes), but weights and states are non-abstract actual values.
    stack = _make_tuple(input_signature)

    # Inputs/outputs of `cond`.
    inputs = inputs_from_stack(stack, self._cond.n_in)
    weights_or_cache_marker, state_or_cache_marker = (
        self._cond.init(inputs, use_cache=True))
    weights.append(weights_or_cache_marker)
    states.append(state_or_cache_marker)
    self._cond._forward_abstract(inputs)
    stack = _make_tuple(outputs_onto_stack([], stack, self._cond.n_in))

    # Inputs/outputs of `true` and `false`.
    for sublayer in [self._true, self._false]:
      inputs = inputs_from_stack(stack, sublayer.n_in)
      weights_or_cache_marker, state_or_cache_marker = (
          sublayer.init(inputs, use_cache=True))
      weights.append(weights_or_cache_marker)
      states.append(state_or_cache_marker)

    self.state = states
    self.weights = weights
    # pylint: enable=protected-access

  def _validate_forward_inputs(self, xs):
    xs = _make_tuple(xs)
    if len(xs) < self.n_in:
      raise ValueError(
          f'Number of inputs ({len(xs)}) to Cond.forward less than n_in '
          f'({self.n_in}).')

  def forward(self, xs):
    """Executes this layer as part of a forward pass through the model.

    Args:
      xs: Tensors of as required by the branches of this conditional.

    Returns:
      Tensors resulting from running the chosen branch.
    """
    # TODO(jaszczur): modify; it's a copy from SkippingSerial
    self._validate_forward_inputs(xs)
    layers_state = self.state
    # Get 3 rngs, one for each layer.
    rngs = _split_rngs(self.rng, 3)

    # Prepare the stack and do some safety checks as in the parent class.
    stack = _make_tuple(xs)
    weights = self.weights
    if len(weights) != 3:
      raise ValueError('number of weights ({}) not equal to 3'
                       .format(len(weights)))
    if len(layers_state) != 3:
      raise ValueError('length of state ({}) not equal to 3'
                       .format(len(layers_state)))

    def true_func(t):
      outputs, new_true_state = self._true.pure_fn(
          t[0][0], t[1][0], t[2][0], t[3][0])
      # t[2][1] is old_false_state which is not changing if true is executed.
      return outputs, (new_true_state, t[2][1])

    def false_func(t):
      if self._identity_false_fun:
        # Memory optimization: we don't need pure_fn call.
        return t[0][1], t[2]
      outputs, new_false_state = self._false.pure_fn(
          t[0][1], t[1][1], t[2][1], t[3][1])
      # t[2][1] is old_true_state, which is not changing if false is executed.
      return outputs, (t[2][0], new_false_state)

    cond_inputs = inputs_from_stack(xs, self._cond.n_in)
    cond_output, s = self._cond.pure_fn(cond_inputs, self.weights[0],
                                        self.state[0], rngs[0], use_cache=True)
    stack = outputs_onto_stack([], stack, self._cond.n_in)
    self._cond.state = s

    outputs, both_states = fastmath.cond(
        cond_output,
        true_func,
        false_func,
        [(stack, stack),
         (self.weights[1], self.weights[2]),
         (self.state[1], self.state[2]),
         (rngs[1], rngs[2])]
    )
    stack = outputs_onto_stack([], stack, self._cond.n_in)

    # We don't know which (`true` or `false`) branch was run, but both of them
    # are adding (n_out) and removing (n_in) the same number of elements of the
    # stack (this was checked in __init__). outputs_onto_stack just uses the
    # layer's n_in, so we can pass either `true` or `false` to it.
    # Note that `outputs` is the actual output of `true` or `false` branch,
    # whichever was run, and we add it to the stack in any case.
    stack = outputs_onto_stack(outputs, stack, self._true.n_in)
    self._true.state = both_states[0]
    self._false.state = both_states[1]
    return _make_singleitem_or_original(stack)


# pylint: disable=invalid-name
def Chunk(layer, chunk_size, pass_unchunkable=True):
  """Executes `layer` using batch chunks of size `chunk_size` to save memory."""
  if chunk_size < 1:
    return layer
  def reshape_to_chunks(x):
    chunk_batch = x.shape[0]
    size = chunk_size
    n_chunks = chunk_batch // size
    if chunk_batch % size != 0:
      if pass_unchunkable:
        n_chunks = 1
        size = chunk_batch
      else:
        raise ValueError(f'Chunk size {size} must divide batch '
                         f'size {chunk_batch}')
    return jnp.reshape(x, [n_chunks, size] + list(x.shape[1:]))
  reshape_to_chunks_layer = base.PureLayer(
      lambda xs: fastmath.nested_map(reshape_to_chunks, xs),
      n_in=layer.n_in, n_out=layer.n_in, name='ReshapeToChunks')
  def reshape_from_chunks(x):
    batch_size = x.shape[0] * x.shape[1]
    return jnp.reshape(x, [batch_size] + list(x.shape[2:]))
  reshape_from_chunks_layer = base.PureLayer(
      lambda xs: fastmath.nested_map(reshape_from_chunks, xs),
      n_in=layer.n_out, n_out=layer.n_out, name='ReshapeFromChunks')
  return Serial(
      reshape_to_chunks_layer,
      Scan(layer, axis=0, n_carry=0, remat=True),
      reshape_from_chunks_layer,
  )


def Branch(*layers, name='Branch'):
  """Combinator that applies a list of layers in parallel to copies of inputs.

  Each layer in the input list is applied to as many inputs from the stack
  as it needs, and their outputs are successively combined on stack.

  For example, suppose one has three layers:

    - F: 1 input, 1 output
    - G: 3 inputs, 1 output
    - H: 2 inputs, 2 outputs (h1, h2)

  Then Branch(F, G, H) will take 3 inputs and give 4 outputs:

    - inputs: a, b, c
    - outputs: F(a), G(a, b, c), h1, h2    where h1, h2 = H(a, b)

  As an important special case, a None argument to Branch acts as if it takes
  one argument, which it leaves unchanged. (It acts as a one-arg no-op.)

  Args:
    *layers: List of layers.
    name: Descriptive name for this layer.

  Returns:
    A branch layer built from the given sublayers.
  """
  if len(layers) == 1:
    return layers[0]
  parallel_layer = Parallel(*layers)
  indices = [list(range(layer.n_in)) for layer in parallel_layer.sublayers]
  return Serial(Select(_deep_flatten(indices)), parallel_layer,
                name=name, sublayers_to_print=layers)


def Residual(*layers, shortcut=None):
  """Wraps a series of layers with a residual connection.

  Args:
    *layers: One or more layers, to be applied in series.
    shortcut: If None (the usual case), the Residual layer computes the
        element-wise sum of the stack-top input with the output of the layer
        series. If specified, the `shortcut` layer applies to a copy of the
        inputs and (elementwise) adds its output to the output from the main
        layer series.

  Returns:
      A layer representing a residual connection paired with a layer series.
  """
  layers = _ensure_flat(layers)
  layer = layers[0] if len(layers) == 1 else Serial(layers)
  # TODO(jonni): Should we require layer.n_out = 1 and shortcut.n_out = 1?
  return Serial(
      Branch(shortcut, layer),
      Add(),  # pylint: disable=no-value-for-parameter
  )


def Select(indices, n_in=None, name=None):
  """Copies, reorders, or deletes stack elements according to `indices`.

  Args:
    indices: A list or tuple of 0-based indices to select elements relative to
        the top of the stack.
    n_in: Number of input elements to pop from the stack, and replace with
        those specified by `indices`. If not specified, its value will be
        calculated as `max(indices) + 1`.
    name: Descriptive name for this layer.

  Returns:
    Tensors, matching the number selected (`n_out = len(indices)`).
    Specifically:

      - n_out = 0: an empty tuple
      - n_out = 1: one tensor (NOT wrapped in a tuple)
      - n_out > 1: a tuple of tensors, with n_out items
  """
  if n_in is None:
    n_in = max(indices) + 1
  if name is None:
    name = f'Select{indices}'.replace(' ', '')

  def select(xs):  # pylint: disable=invalid-name
    if not isinstance(xs, (tuple, list)):
      xs = (xs,)
    selected = tuple(xs[i] for i in indices)
    return selected[0] if len(selected) == 1 else selected

  return base.PureLayer(select, n_in=n_in, n_out=len(indices), name=name)


def Drop():
  """Drops the top stack element."""
  return Fn('Drop', lambda x: (), n_out=0)


def Dup():
  """Duplicates (copies) the top element on the data stack."""
  return Fn('Dup', lambda x: (x, x), n_out=2)


def Swap():
  """Swaps the top two stack elements."""
  return Fn('Swap', lambda x0, x1: (x1, x0), n_out=2)


def SerialWithSideOutputs(layers, n_side_outputs=1):
  """Serial layer with side outputs.

  This layer makes it easier to manage the stack when layers have side outputs.

  In the simplest case of layers with n_in=1, n_out=2 and with
  n_side_outputs=1, this layer runs the following computation on x::

    side_outputs = []
    for i in range(len(layers)):
      x, side_output = layers[i](x)
      side_outputs.append(side_output)
    return [x] + side_outputs

  In the general case of layers with variable n_in and n_out and
  n_side_outputs being a list of N integers, it does the following::

    side_outputs = []
    for i in range(N):
      res = layer[i](cur_stack)  # remove n_in from stack
      cur_stack.append(res[:n_side_outputs[i]])  # put back some on stack
      side_outputs.extend(res[n_side_outputs:])
    return cur_stack + side_outputs

  Args:
    layers: a list of layers to execute
    n_side_outputs: an int or a list of ints, how many outputs of each layer
        to put aside

  Returns:
    A layer that performs the above computation.
  """
  if isinstance(n_side_outputs, int):
    n_side_outputs = [n_side_outputs] * len(layers)

  # Calculate the n_in for this layer.
  running_max = 0
  running_total = 0
  for layer, n_side_output in zip(layers, n_side_outputs):
    running_total += layer.n_in
    running_max = max(running_max, running_total)
    running_total -= layer.n_out - n_side_output
  n_in = running_max

  # Create the list of layers to run serially.
  cur_stack_size = n_in
  serial_layers = []
  for layer, n_side_output in zip(layers, n_side_outputs):
    serial_layers.append(layer)
    cur_stack_size += layer.n_out - layer.n_in
    # Indices to move n_side_outputs to the back of the stack.
    # Don't touch first n_out - n_side_outputs.
    move_back_indices = list(range(layer.n_out - n_side_output))
    # Then comes the rest of the stack that we're not moving.
    move_back_indices += [i + layer.n_out
                          for i in range(cur_stack_size - layer.n_out)]
    # Finally the indices we move.
    move_back_indices += [i + layer.n_out - n_side_output
                          for i in range(n_side_output)]
    # Swap them on stack.
    serial_layers.append(Select(move_back_indices))

  return Serial(serial_layers)


def FlattenList():
  """Flatten lists."""
  # TODO(jonni): Consider renaming layer to DeepFlatten.
  return Fn('FlattenList', lambda x: tuple(_deep_flatten(x)))


def Add():
  """Adds two tensors."""
  return Fn('Add', lambda x0, x1: x0 + x1)


def SubtractTop():
  """Subtracts the first tensor from the second."""
  return Fn('SubtractTop', lambda x0, x1: x1 - x0)


def Multiply():
  """Multiplies two tensors."""
  return Fn('Multiply', lambda x0, x1: x0 * x1)


def Gate():
  """Returns a gating layer on a (memory, gate, candidate) tuple.

  Final update is memory * gate + (1 - gate) * candidate

  This gating equation may also be referred to as Highway Network.
  Highway Networks: https://arxiv.org/abs/1505.00387
  """
  return Fn('Gate', lambda m, g, c: g * m + (1.0 - g) * c)


class Cache(base.Layer):
  """Applies a layer on the first run and returns the outputs on next calls."""

  def __init__(self, layer):
    super().__init__(n_in=layer.n_in, n_out=layer.n_out)
    self._sublayers = [layer]

  @property
  def sublayer(self):
    """Returns the unique sublayer managed by this layer."""
    return self._sublayers[0]

  @property
  def state(self):
    """Returns a tuple containing this layer's state; may be empty."""
    return self._state

  @state.setter
  def state(self, state):
    """Recursively sets state on this layer and all sublayers."""
    if isinstance(state, dict) and state == base.GET_STATE_FROM_CACHE:
      return
    self._state = state
    self.sublayer.state = state[1]

  def init_weights_and_state(self, input_signature):
    """Initializes weights and state for inputs with the given signature."""
    weights, layer_state = self.sublayer.init(input_signature, use_cache=True)
    self.state = ((), layer_state)
    self._weights = (weights,)

  def forward(self, inputs):
    """Executes this layer as part of a forward pass through the model.

    Args:
      inputs: Tensors required by the sublayer.

    Returns:
      Tensors resulting from running the sublayer the first time.
    """
    state, weights = self.state, self.weights[0]
    if state[0] is ():  # pylint: disable=literal-comparison
      res, layer_state = self.sublayer.pure_fn(
          inputs, weights, state[1], self.rng)
      self.state = (res, layer_state)
      return res
    else:
      return state[0]


class BatchLeadingAxes(base.Layer):
  """Applies a layer after flattening all but n_last_axes_to_keep to batch.

  This can be used to make layers accept an arbitrary number of leading
  axes (dimensions) as batch. For example, a Convolution layer may normally
  only operate on tensors of shape [B, W, H, C]. In this case, the layer

      BatchLeadingAxes(Convolution(), n_last_axes_to_keep=3)

  will operate on any tensor [..., W, H, C] and treat the leading axes as batch.
  """

  def __init__(self, layer, n_last_axes_to_keep=1):
    if layer.n_out != 1:
      raise ValueError('BatchLeadingAxes currently only works for layers with '
                       f'n_out = 1, got {layer.n_out}.')
    super().__init__(n_in=layer.n_in, n_out=layer.n_out)
    self._sublayers = [layer]
    self._n_last_axes_to_keep = n_last_axes_to_keep
    self._weights = (None,)
    self._state = (None,)

  @property
  def sublayer(self):
    """Returns the unique sublayer managed by this layer."""
    return self._sublayers[0]

  def forward(self, inputs):
    """Executes this layer as part of a forward pass through the model."""
    if self._n_in == 1:
      inputs = [inputs]
    new_inputs = []
    for old_input in inputs:
      batched_axes_shape = list(old_input.shape[:-self._n_last_axes_to_keep])
      batched_shape = [-1] + list(old_input.shape[-self._n_last_axes_to_keep:])
      new_inputs.append(jnp.reshape(old_input, batched_shape))
    new_inputs = tuple(new_inputs)
    if self._n_in == 1:
      new_inputs = new_inputs[0]
    res, layer_state = self.sublayer.pure_fn(
        new_inputs, self.weights[0], self.state[0], self.rng)
    self.state = (layer_state,)
    return jnp.reshape(res, batched_axes_shape + list(res.shape[1:]))

  def init_weights_and_state(self, input_signature):
    """Initializes weights and state for inputs with the given signature."""
    if self._n_in == 1 and not isinstance(input_signature, (list, tuple)):
      input_signature = (input_signature,)
    batched_signature = []
    for sub_input_signature in input_signature:
      batched_size = 1
      for d in sub_input_signature.shape[:-self._n_last_axes_to_keep]:
        batched_size *= d
      batched_shape = [batched_size] + list(
          sub_input_signature.shape[-self._n_last_axes_to_keep:])
      batched_signature.append(ShapeDtype(batched_shape,
                                          sub_input_signature.dtype))
    if self._n_in == 1:
      batched_signature = batched_signature[0]
    weights, layer_state = self.sublayer.init(batched_signature, use_cache=True)
    self.state = (layer_state,)
    self.weights = (weights,)


def Bidirectional(forward_layer, axis=1, merge_layer=Concatenate()):
  """Bidirectional combinator for RNNs.

  Args:
    forward_layer: A layer, such as `trax.layers.LSTM` or `trax.layers.GRU`.
    axis: a time axis of the inputs. Default value is `1`.
    merge_layer: A combinator used to combine outputs of the forward
      and backward RNNs. Default value is 'trax.layers.Concatenate'.

  Example:
      Bidirectional(RNN(n_units=8))

  Returns:
    The Bidirectional combinator for RNNs.
  """
  backward_layer = copy.deepcopy(forward_layer)
  flip = base.Fn('_FlipAlongTimeAxis', lambda x: jnp.flip(x, axis=axis))
  backward = Serial(
      flip,
      backward_layer,
      flip,
  )

  return Serial(
      Branch(forward_layer, backward),
      merge_layer,
  )


# All module-private helper functions are below.
# pylint: disable=invalid-name


def _deep_flatten(items):
  """Returns a list of objects, flattening sublists/subtuples along the way.

  Example: _deep_flatten([1, (2, 3, (4, 5), [6, 7]), [[[8]]]]) would return
  the list [1, 2, 3, 4, 5, 6, 7, 8].

  Args:
    items: An iterable. If elements of this iterable are lists or tuples, they
        will be (recursively) flattened until non-list non-tuple objects are
        reached.

  Returns:
    A list of non-list, non-tuple objects.
  """
  def _flat_gen(xs):
    for x in xs:
      if isinstance(x, (list, tuple)):
        for y in _flat_gen(x):
          yield y
      else:
        yield x
  return list(_flat_gen(items))


def _ensure_sublayers(layers):
  """Ensures that elements in a layer list are layers.

  Args:
    layers: A tuple or list whose elements can each be a layer, tuple, or list,
        and so on recursively.

  Returns:
    An analogous collection of layers in which embedded layer lists are
    wrapped in Serial layer instances.
  """
  if not layers:  # None or an empty list can signal a no-op.
    return Serial(None)  # no-op, but still handles shapes and initialization
  elif isinstance(layers, (list, tuple)):
    sublayers_not_lists = []
    for layer in layers:
      sublayers_not_lists.append(
          Serial(layer) if isinstance(layer, (list, tuple)) else layer)
    return sublayers_not_lists
  else:
    raise TypeError(type(layers))


def _split_rngs(rng, n_copies):
  if rng is None:
    return (None,) * n_copies
  return fastmath.random.split(rng, n_copies)


def inputs_from_stack(stack, n):
  """Returns n inputs from stack."""
  stack = _make_tuple(stack)
  return _make_singleitem_or_original(stack[:n])


def outputs_onto_stack(outputs, stack, n):
  """"Returns the new stack after removing n items and pushing outputs there."""
  outputs = _make_tuple(outputs)
  stack = _make_tuple(stack)
  return _make_singleitem_or_original(outputs + stack[n:])


def _make_tuple(xs):
  """Returns a tuple from a list, a tuple, or a single element."""
  if isinstance(xs, (list, tuple)):
    return tuple(xs)
  else:
    return (xs,)


def _make_singleitem_or_original(xs):
  """Returns a single element if possible, or the original list/tuple if not."""
  if isinstance(xs, (list, tuple)) and len(xs) == 1:
    return xs[0]
  else:
    return xs


def _shape_without_axis(x, axis):
  return x.shape[:axis] + x.shape[axis + 1:]


def _ensure_flat(layers):
  """Ensures that layers is a single flat list of Layer instances."""
  if len(layers) == 1 and layers[0] is None:
    layers = ()
  else:
    layers = _deep_flatten(layers)
  for obj in layers:
    if not isinstance(obj, base.Layer):
      raise ValueError(
          f'Found nonlayer object ({obj}) in layers: {layers}')
  return layers



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for ReZero models."""

from absl.testing import absltest
import numpy as np

from trax import layers as tl
from trax import shapes
from trax.models.research import rezero


class ResidualZeroTest(absltest.TestCase):

  def test_residual_layer_forward(self):
    """Tests that the forward pass runs and returns the expected shape."""
    model = rezero.ResidualZero(tl.Dense(5))
    x = [np.arange(5).astype(np.float32)]
    _, _ = model.init(shapes.signature(x))
    y = model(x)
    self.assertEqual(y.tolist(), [0., 1., 2., 3., 4.])


class ReZeroTransformerLMTest(absltest.TestCase):

  def test_rezero_lm_forward_shape(self):
    """Tests that the forward pass runs and returns the expected shape."""
    vocab_size = 16
    model = rezero.ReZeroTransformerLM(
        vocab_size, d_model=32, d_ff=64, n_layers=2, n_heads=2, max_len=16)
    xs = [np.ones((1, 8)).astype(np.int32),
          np.ones((1, 8)).astype(np.int32)]
    _, _ = model.init(shapes.signature(xs))
    ys = model(xs)
    self.assertEqual([y.shape for y in ys], [(1, 8, 16), (1, 8)])


class ReZeroTransformerTest(absltest.TestCase):

  def test_rezero_forward_shape(self):
    """Tests that the forward pass runs and returns the expected shape."""
    vocab_size = 16
    model = rezero.ReZeroTransformer(
        vocab_size, d_model=32, d_ff=64, n_encoder_layers=2, n_decoder_layers=2,
        n_heads=2, max_len=16)
    xs = [np.ones((1, 8)).astype(np.int32),
          np.ones((1, 8)).astype(np.int32)]
    _, _ = model.init(shapes.signature(xs))
    ys = model(xs)
    self.assertEqual([y.shape for y in ys], [(1, 8, 16), (1, 8)])


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""RNNs (recursive neural networks)."""

from trax import layers as tl
from trax.fastmath import numpy as jnp


def RNNLM(vocab_size,
          d_model=512,
          n_layers=2,
          rnn_cell=tl.LSTMCell,
          rnn_cell_d_state_multiplier=2,
          dropout=0.1,
          mode='train'):
  """Returns an RNN language model.

  This model performs autoregressive language modeling:

    - input: rank 2 tensor representing a batch of text strings via token IDs
      plus padding markers; shape is (batch_size, sequence_length). The tensor
      elements are integers in `range(vocab_size)`, and `0` values mark padding
      positions.

    - output: rank 3 tensor representing a batch of log-probability
      distributions for each sequence position over possible token IDs;
      shape is (batch_size, sequence_length, `vocab_size`).

  Args:
    vocab_size: Input vocabulary size -- each element of the input tensor
        should be an integer in `range(vocab_size)`. These integers typically
        represent token IDs from a vocabulary-based tokenizer.
    d_model: Embedding depth throughout the model.
    n_layers: Number of RNN layers.
    rnn_cell: Type of RNN cell; must be a subclass of `Layer`.
    rnn_cell_d_state_multiplier: Multiplier for feature depth of RNN cell
        state.
    dropout: Stochastic rate (probability) for dropping an activation value
        when applying dropout.
    mode: If `'predict'`, use fast inference; if `'train'` apply dropout.

  Returns:
    An RNN language model as a layer that maps from a tensor of tokens
    to activations over a vocab set.
  """

  if n_layers != 2:  # TODO(jonni): Remove n_layers arg, if it can't vary?
    raise ValueError(f'Number of layers must be set to 2; instead got'
                     f' {n_layers}.')

  def MultiRNNCell():
    """Multi-layer RNN cell."""
    return tl.Serial(
        tl.Parallel([], tl.Split(n_items=n_layers)),
        tl.SerialWithSideOutputs(
            [rnn_cell(n_units=d_model) for _ in range(n_layers)]),
        tl.Parallel([], tl.Concatenate(n_items=n_layers))
    )

  zero_state = tl.MakeZeroState(  # pylint: disable=no-value-for-parameter
      depth_multiplier=n_layers * rnn_cell_d_state_multiplier
  )

  return tl.Serial(
      tl.ShiftRight(mode=mode),
      tl.Embedding(vocab_size, d_model),
      tl.Dropout(rate=dropout, mode=mode),
      tl.Branch([], zero_state),
      tl.Scan(MultiRNNCell(), axis=1, mode=mode),
      tl.Select([0], n_in=2),  # Drop RNN state.
      tl.Dense(vocab_size),
  )


def GRULM(vocab_size=256,
          d_model=512,
          n_layers=1,
          mode='train'):
  """Returns a GRU (gated recurrent unit) language model.

  This model performs autoregressive language modeling:

    - input: rank 2 tensor representing a batch of text strings via token IDs
      plus padding markers; shape is (batch_size, sequence_length). The tensor
      elements are integers in `range(vocab_size)`, and `0` values mark padding
      positions.

    - output: rank 3 tensor representing a batch of log-probability
      distributions for each sequence position over possible token IDs;
      shape is (batch_size, sequence_length, `vocab_size`).

  Args:
    vocab_size: Input vocabulary size -- each element of the input tensor
        should be an integer in `range(vocab_size)`. These integers typically
        represent token IDs from a vocabulary-based tokenizer.
    d_model: Embedding depth throughout the model.
    n_layers: Number of GRU layers.
    mode: If `'predict'`, use fast inference (and omit the right shift).

  Returns:
    A GRU language model as a layer that maps from a tensor of tokens
    to activations over a vocab set.
  """
  return tl.Serial(
      tl.ShiftRight(mode=mode),
      tl.Embedding(vocab_size, d_model),
      [tl.GRU(d_model, mode=mode) for _ in range(n_layers)],
      tl.Dense(vocab_size),
  )


# TODO(jonni): Decide names (here and Transformer): input/source, output/target
# TODO(jonni): Align with Transfomer: (attention-)dropout, n-(attention-)heads
def LSTMSeq2SeqAttn(input_vocab_size=256,
                    target_vocab_size=256,
                    d_model=512,
                    n_encoder_layers=2,
                    n_decoder_layers=2,
                    n_attention_heads=1,
                    attention_dropout=0.0,
                    mode='train'):
  """Returns an LSTM sequence-to-sequence model with attention.

  This model is an encoder-decoder that performs tokenized string-to-string
  ("source"-to-"target") transduction:

    - inputs (2):

        - source: rank 2 tensor representing a batch of text strings via token
          IDs plus padding markers; shape is (batch_size, sequence_length). The
          tensor elements are integers in `range(input_vocab_size)`, and `0`
          values mark padding positions.

        - target: rank 2 tensor representing a batch of text strings via token
          IDs plus padding markers; shape is (batch_size, sequence_length). The
          tensor elements are integers in `range(output_vocab_size)`, and `0`
          values mark padding positions.

    - output: rank 3 tensor representing a batch of log-probability
      distributions for each sequence position over possible token IDs;
      shape is (batch_size, sequence_length, `vocab_size`).

  An example use would be to translate (tokenized) sentences from English to
  German.

  The model works as follows:

  * Input encoder runs on the input tokens and creates activations that
    are used as both keys and values in attention.
  * Pre-attention decoder runs on the targets and creates
    activations that are used as queries in attention.
  * Attention runs on the queries, keys and values masking out input padding.
  * Decoder runs on the result, followed by a cross-entropy loss.

  Args:
    input_vocab_size: Input vocabulary size -- each element of the input tensor
        should be an integer in `range(vocab_size)`. These integers typically
        represent token IDs from a vocabulary-based tokenizer.
    target_vocab_size: Target vocabulary size.
    d_model: Final dimension of tensors at most points in the model, including
        the initial embedding output.
    n_encoder_layers: Number of LSTM layers in the encoder.
    n_decoder_layers: Number of LSTM layers in the decoder after attention.
    n_attention_heads: Number of attention heads.
    attention_dropout: Stochastic rate (probability) for dropping an activation
        value when applying dropout within an attention block.
    mode: If `'predict'`, use fast inference. If `'train'`, each attention block
        will include dropout; else, it will pass all values through unaltered.

  Returns:
    An LSTM sequence-to-sequence model as a layer that maps from a
    source-target tokenized text pair to activations over a vocab set.
  """
  input_encoder = tl.Serial(
      tl.Embedding(input_vocab_size, d_model),
      [tl.LSTM(d_model) for _ in range(n_encoder_layers)],
  )

  pre_attention_decoder = tl.Serial(
      tl.ShiftRight(mode=mode),
      tl.Embedding(target_vocab_size, d_model),
      tl.LSTM(d_model, mode=mode),
  )

  def PrepareAttentionInputs():
    """Layer that prepares queries, keys, values and mask for attention."""
    def F(encoder_activations, decoder_activations, input_tokens):
      keys = values = encoder_activations
      queries = decoder_activations
      # Mask is 1 where inputs are not padding (0) and 0 where they are padding.
      mask = (input_tokens != 0)
      # We need to add axes to the mask for attention heads and decoder length.
      mask = jnp.reshape(mask, (mask.shape[0], 1, 1, mask.shape[1]))
      # Broadcast so mask is [batch, 1 for heads, decoder-len, encoder-len].
      mask = mask + jnp.zeros((1, 1, decoder_activations.shape[1], 1))
      mask = mask.astype(jnp.float32)
      return queries, keys, values, mask
    return tl.Fn('PrepareAttentionInputs', F, n_out=4)

  return tl.Serial(              # in-toks, target-toks
      tl.Select([0, 1, 0, 1]),   # in-toks, target-toks, in-toks, target-toks
      tl.Parallel(input_encoder, pre_attention_decoder),
      PrepareAttentionInputs(),  # q, k, v, mask, target-toks
      tl.Residual(
          tl.AttentionQKV(d_model, n_heads=n_attention_heads,
                          dropout=attention_dropout, mode=mode,
                          cache_KV_in_predict=True)
      ),                         # decoder-vecs, mask, target-toks
      tl.Select([0, 2]),         # decoder-vecs, target-toks
      [tl.LSTM(d_model, mode=mode) for _ in range(n_decoder_layers)],
      tl.Dense(target_vocab_size),
      tl.LogSoftmax()
  )



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for trax.rl.serialization_utils."""
import functools

from absl.testing import absltest
from absl.testing import parameterized
import gin
import gym
from jax import numpy as jnp
import numpy as np
from trax import models as trax_models
from trax import shapes
from trax import test_utils
from trax.data import inputs as trax_input
from trax.layers import base as layers_base
from trax.models import transformer
from trax.rl import serialization_utils
from trax.rl import space_serializer
from trax.supervised import trainer_lib


# pylint: disable=invalid-name
def TestModel(extra_dim, mode='train'):
  """Dummy sequence model for testing."""
  del mode
  def f(inputs):
    # Cast the input to float32 - this is for simulating discrete-input models.
    inputs = inputs.astype(np.float32)
    # Add an extra dimension if requested, e.g. the logit dimension for output
    # symbols.
    if extra_dim is not None:
      return jnp.broadcast_to(inputs[:, :, None], inputs.shape + (extra_dim,))
    else:
      return inputs
  return layers_base.Fn('TestModel', f)
  # pylint: enable=invalid-name


def signal_inputs(seq_len, batch_size, depth=1):
  def stream_fn(num_devices):
    del num_devices
    while True:
      x = np.random.uniform(size=(batch_size, seq_len, depth))
      y = np.random.uniform(size=(batch_size, seq_len, depth))
      mask = np.ones_like(x).astype(np.float32)
      yield (x, y, x, mask)

  return trax_input.Inputs(
      train_stream=stream_fn,
      eval_stream=stream_fn,
  )


class SerializationTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._serializer = space_serializer.create(
        gym.spaces.Discrete(2), vocab_size=2
    )
    self._repr_length = 100
    self._serialization_utils_kwargs = {
        'observation_serializer': self._serializer,
        'action_serializer': self._serializer,
        'representation_length': self._repr_length,
    }
    test_utils.ensure_flag('test_tmpdir')

  def test_serialized_model_discrete(self):
    vocab_size = 3
    obs = np.array([[[0, 1], [1, 1], [1, 0], [0, 0]]])
    act = np.array([[1, 0, 0]])
    mask = np.array([[1, 1, 1, 0]])

    test_model_inputs = []

    # pylint: disable=invalid-name
    def TestModelSavingInputs(mode):
      del mode
      def f(inputs):
        # Save the inputs for a later check.
        test_model_inputs.append(inputs)
        # Change type to np.float32 and add the logit dimension.
        return jnp.broadcast_to(
            inputs.astype(np.float32)[:, :, None], inputs.shape + (vocab_size,)
        )
      return layers_base.Fn('TestModelSavingInputs', f)
      # pylint: enable=invalid-name

    obs_serializer = space_serializer.create(
        gym.spaces.MultiDiscrete([2, 2]), vocab_size=vocab_size
    )
    act_serializer = space_serializer.create(
        gym.spaces.Discrete(2), vocab_size=vocab_size
    )
    serialized_model = serialization_utils.SerializedModel(
        TestModelSavingInputs,  # pylint: disable=no-value-for-parameter
        observation_serializer=obs_serializer,
        action_serializer=act_serializer,
        significance_decay=0.9,
    )

    example = (obs, act, obs, mask)
    serialized_model.init(shapes.signature(example))

    (obs_logits, obs_repr, weights) = serialized_model(example)
    # Check that the model has been called with the correct input.
    np.testing.assert_array_equal(
        # The model is called multiple times for determining shapes etc.
        # Check the last saved input - that should be the actual concrete array
        # calculated during the forward pass.
        test_model_inputs[-1],
        # Should be serialized observations and actions interleaved.
        [[0, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0]],
    )
    # Check the output shape.
    self.assertEqual(obs_logits.shape, obs_repr.shape + (vocab_size,))
    # Check that obs_logits are the same as obs_repr, just broadcasted over the
    # logit dimension.
    np.testing.assert_array_equal(np.min(obs_logits, axis=-1), obs_repr)
    np.testing.assert_array_equal(np.max(obs_logits, axis=-1), obs_repr)
    # Check that the observations are correct.
    np.testing.assert_array_equal(obs_repr, obs)
    # Check weights.
    np.testing.assert_array_equal(
        weights,
        [[[1., 1.], [1., 1.], [1., 1.], [0., 0.]]],
    )

  def test_train_model_with_serialization(self):
    # Serializer handles discretization of the data.
    precision = 2
    number_of_time_series = 2
    vocab_size = 16
    srl = space_serializer.BoxSpaceSerializer(
        space=gym.spaces.Box(shape=(number_of_time_series,),
                             low=0.0, high=16.0),
        vocab_size=vocab_size,
        precision=precision,
    )

    def model(mode):
      del mode
      return serialization_utils.SerializedModel(
          functools.partial(
              trax_models.TransformerLM,
              vocab_size=vocab_size,
              d_model=16,
              d_ff=8,
              n_layers=1,
              n_heads=1,
          ),
          observation_serializer=srl,
          action_serializer=srl,
          significance_decay=0.9,
      )

    output_dir = self.create_tempdir().full_path
    state = trainer_lib.train(
        output_dir=output_dir,
        model=model,
        inputs=functools.partial(signal_inputs, seq_len=5,
                                 batch_size=64, depth=number_of_time_series),
        steps=2)
    self.assertEqual(2, state.step)

  def test_serialized_model_continuous(self):
    precision = 3
    gin.bind_parameter('BoxSpaceSerializer.precision', precision)

    vocab_size = 32
    obs = np.array([[[1.5, 2], [-0.3, 1.23], [0.84, 0.07], [0, 0]]])
    act = np.array([[0, 1, 0]])
    mask = np.array([[1, 1, 1, 0]])

    obs_serializer = space_serializer.create(
        gym.spaces.Box(shape=(2,), low=-2, high=2), vocab_size=vocab_size
    )
    act_serializer = space_serializer.create(
        gym.spaces.Discrete(2), vocab_size=vocab_size
    )
    serialized_model = serialization_utils.SerializedModel(
        functools.partial(TestModel, extra_dim=vocab_size),
        observation_serializer=obs_serializer,
        action_serializer=act_serializer,
        significance_decay=0.9,
    )

    example = (obs, act, obs, mask)
    serialized_model.init(shapes.signature(example))

    (obs_logits, obs_repr, weights) = serialized_model(example)
    self.assertEqual(obs_logits.shape, obs_repr.shape + (vocab_size,))
    self.assertEqual(
        obs_repr.shape, (1, obs.shape[1], obs.shape[2] * precision)
    )
    self.assertEqual(obs_repr.shape, weights.shape)

  def test_serialized_model_extracts_seq_model_weights_and_state(self):
    vocab_size = 3

    seq_model_fn = functools.partial(
        transformer.TransformerLM,
        vocab_size=vocab_size,
        d_model=2,
        d_ff=2,
        n_layers=0,
    )
    seq_model = seq_model_fn(mode='eval')
    obs_serializer = space_serializer.create(
        gym.spaces.Discrete(2), vocab_size=vocab_size
    )
    act_serializer = space_serializer.create(
        gym.spaces.Discrete(2), vocab_size=vocab_size
    )
    serialized_model = serialization_utils.SerializedModel(
        seq_model_fn,
        observation_serializer=obs_serializer,
        action_serializer=act_serializer,
        significance_decay=0.9,
    )

    obs_sig = shapes.ShapeDtype((1, 2))
    act_sig = shapes.ShapeDtype((1, 1))
    serialized_model.init(input_signature=(obs_sig, act_sig, obs_sig, obs_sig))
    seq_model.weights = serialized_model.seq_model_weights
    seq_model.state = serialized_model.seq_model_state
    # Run the model to check if the weights and state have correct structure.
    seq_model(jnp.array([[0]]))

  @parameterized.named_parameters(('raw', None), ('serialized', 32))
  def test_wrapped_policy_continuous(self, vocab_size):
    precision = 3
    n_controls = 2
    n_actions = 4
    gin.bind_parameter('BoxSpaceSerializer.precision', precision)

    obs = np.array([[[1.5, 2], [-0.3, 1.23], [0.84, 0.07], [0.01, 0.66]]])
    act = np.array([[[0, 1], [2, 0], [1, 3]]])

    wrapped_policy = serialization_utils.wrap_policy(
        TestModel(extra_dim=vocab_size),  # pylint: disable=no-value-for-parameter
        observation_space=gym.spaces.Box(shape=(2,), low=-2, high=2),
        action_space=gym.spaces.MultiDiscrete([n_actions] * n_controls),
        vocab_size=vocab_size,
    )

    example = (obs, act)
    wrapped_policy.init(shapes.signature(example))
    (act_logits, values) = wrapped_policy(example)
    self.assertEqual(act_logits.shape, obs.shape[:2] + (n_controls, n_actions))
    self.assertEqual(values.shape, obs.shape[:2])

  def test_analyzes_discrete_action_space(self):
    space = gym.spaces.Discrete(n=5)
    (n_controls, n_actions) = serialization_utils.analyze_action_space(space)
    self.assertEqual(n_controls, 1)
    self.assertEqual(n_actions, 5)

  def test_analyzes_multi_discrete_action_space_with_equal_categories(self):
    space = gym.spaces.MultiDiscrete(nvec=(3, 3))
    (n_controls, n_actions) = serialization_utils.analyze_action_space(space)
    self.assertEqual(n_controls, 2)
    self.assertEqual(n_actions, 3)

  def test_doesnt_analyze_multi_disccrete_action_space_with_inequal_categories(
      self
  ):
    space = gym.spaces.MultiDiscrete(nvec=(2, 3))
    with self.assertRaises(AssertionError):
      serialization_utils.analyze_action_space(space)

  def test_doesnt_analyze_box_action_space(self):
    space = gym.spaces.Box(shape=(2, 3), low=0, high=1)
    with self.assertRaises(AssertionError):
      serialization_utils.analyze_action_space(space)


if __name__ == '__main__':
  absltest.main()
